import argparse
import logging
import time
from pathlib import Path
from typing import Any
from typing import Optional

import numpy as np
import numpy.typing as npt
import polars as pl
import pyarrow as pa
import pyarrow.parquet as pq
import torch
from torch.utils.data import DataLoader

from birder.common import cli
from birder.common import fs_ops
from birder.common import lib
from birder.conf import settings
from birder.data.dataloader.webdataset import make_wds_loader
from birder.data.datasets.directory import make_image_dataset
from birder.data.datasets.webdataset import make_wds_dataset
from birder.data.datasets.webdataset import prepare_wds_args
from birder.data.datasets.webdataset import wds_args_from_info
from birder.data.transforms.classification import inference_preset
from birder.inference.classification import infer_dataloader_iter
from birder.inference.data_parallel import InferenceDataParallel
from birder.results.classification import Results
from birder.results.classification import SparseResults
from birder.results.gui import show_top_k

logger = logging.getLogger(__name__)


def _init_array_parquet_writer(
    metadata_columns: list[str], array_name: str, array_len: int, path: Path
) -> pq.ParquetWriter:
    embeddings_schema = pa.schema(
        [pa.field(name, pa.string()) for name in metadata_columns]
        + [pa.field(array_name, pa.list_(pa.float32(), list_size=array_len))]
    )
    return pq.ParquetWriter(path, embeddings_schema)


def _init_parquet_writer(metadata_columns: list[str], label_names: list[str], path: Path) -> pq.ParquetWriter:
    logits_schema = pa.schema(
        [pa.field(name, pa.string()) for name in metadata_columns]
        + [pa.field(name, pa.float32()) for name in label_names]
    )
    return pq.ParquetWriter(path, logits_schema)


def save_embeddings_parquet(
    writer: pq.ParquetWriter, sample_paths: list[str], embedding_list: list[npt.NDArray[np.float32]]
) -> None:
    logger.info(f"Writing embeddings at {writer.where}")
    embeddings = np.concatenate(embedding_list, axis=0)
    table = pa.Table.from_pydict(
        {
            "sample": pa.array(sample_paths, type=pa.string()),
            "embedding": pa.array(embeddings.tolist(), type=pa.list_(pa.float32())),
        },
        schema=writer.schema,
    )
    writer.write_table(table)


def save_logits_parquet(writer: pq.ParquetWriter, sample_paths: list[str], logits: npt.NDArray[np.float32]) -> None:
    logger.info(f"Writing logits at {writer.where}")
    data = {"sample": pa.array(sample_paths, type=pa.string())}
    for i, field_name in enumerate(writer.schema.names[1:]):
        data[field_name] = pa.array(logits[:, i], type=pa.float32())

    table = pa.Table.from_pydict(data, schema=writer.schema)
    writer.write_table(table)


def save_output_parquet(
    writer: pq.ParquetWriter, sample_paths: list[str], label_names: list[str], outs: npt.NDArray[np.float32]
) -> None:
    logger.info(f"Writing outputs at {writer.where}")
    data = {
        "sample": pa.array(sample_paths, type=pa.string()),
        "prediction": pa.array(np.array(label_names)[outs.argmax(axis=1)], type=pa.string()),
    }
    for i, field_name in enumerate(writer.schema.names[2:]):
        data[field_name] = pa.array(outs[:, i], type=pa.float32())

    table = pa.Table.from_pydict(data, schema=writer.schema)
    writer.write_table(table)


def _save_dataframe(path: Path, df: pl.DataFrame, append: bool, data_type: str) -> None:
    if append is False:
        logger.info(f"Saving {data_type} at {path}")
        df.write_csv(path)
    else:
        logger.info(f"Adding {data_type} to {path}")
        with open(path, "a", encoding="utf-8") as handle:
            df.write_csv(handle, include_header=False)


def save_embeddings(
    path: Path, sample_paths: list[str], embedding_list: list[npt.NDArray[np.float32]], append: bool
) -> None:
    embeddings = np.concatenate(embedding_list, axis=0)
    embeddings_df = pl.DataFrame(
        {
            "sample": sample_paths,
            **{f"{i}": embeddings[:, i] for i in range(embeddings.shape[-1])},
        }
    )
    embeddings_df = embeddings_df.sort("sample", descending=False)
    _save_dataframe(path, embeddings_df, append, data_type="embeddings")


def save_logits(
    path: Path, sample_paths: list[str], label_names: list[str], logits: npt.NDArray[np.float32], append: bool
) -> None:
    logits_df = pl.DataFrame(
        {
            "sample": sample_paths,
            **{name: logits[:, i] for i, name in enumerate(label_names)},
        }
    )
    logits_df = logits_df.sort("sample", descending=False)
    _save_dataframe(path, logits_df, append, data_type="logits")


def save_output(
    path: Path, sample_paths: list[str], label_names: list[str], outs: npt.NDArray[np.float32], append: bool
) -> None:
    output_df = pl.DataFrame(
        {
            "sample": sample_paths,
            "prediction": np.array(label_names)[outs.argmax(axis=1)],
            **{name: outs[:, i] for i, name in enumerate(label_names)},
        }
    )
    output_df = output_df.sort("sample", descending=False)
    _save_dataframe(path, output_df, append, data_type="output")


def handle_show_flags(
    args: argparse.Namespace,
    img_path: str,
    prob: npt.NDArray[np.float32],
    label: int,
    class_to_idx: dict[str, int],
) -> None:
    # Show prediction
    if args.show is True:
        show_top_k(img_path, prob, class_to_idx, label)
    elif args.show_top_below is not None and args.show_top_below > prob.max():
        show_top_k(img_path, prob, class_to_idx, label)

    # Show mistake (if label exists)
    if label != settings.NO_LABEL:
        if args.show_target_below is not None and args.show_target_below > prob[label]:
            show_top_k(img_path, prob, class_to_idx, label)

        elif args.show_mistakes is True:
            if label != np.argmax(prob):
                show_top_k(img_path, prob, class_to_idx, label)

        elif args.show_out_of_k is True:
            if label not in np.argsort(prob)[::-1][0 : settings.TOP_K]:
                show_top_k(img_path, prob, class_to_idx, label)

        elif args.show_class is not None:
            idx_to_class = dict(zip(class_to_idx.values(), class_to_idx.keys()))
            if args.show_class == idx_to_class[np.argmax(prob)]:  # type: ignore
                show_top_k(img_path, prob, class_to_idx, label)


# pylint: disable=too-many-locals,too-many-branches,too-many-statements
def predict(args: argparse.Namespace) -> None:
    if args.gpu is True:
        device = torch.device("cuda")
    elif args.mps is True:
        device = torch.device("mps")
    else:
        device = torch.device("cpu")

    if args.parallel is True and torch.cuda.device_count() > 1:
        logger.info(f"Using {torch.cuda.device_count()} {device} devices")
    else:
        if args.gpu_id is not None:
            torch.cuda.set_device(args.gpu_id)

        logger.info(f"Using device {device}")

    model_dtype: torch.dtype = getattr(torch, args.model_dtype)
    if args.amp_dtype is None:
        amp_dtype = torch.get_autocast_dtype(device.type)
    else:
        amp_dtype = getattr(torch, args.amp_dtype)

    network_name = lib.get_network_name(args.network, tag=args.tag)
    (net, (class_to_idx, signature, rgb_stats, *_)) = fs_ops.load_model(
        device,
        args.network,
        config=args.model_config,
        tag=args.tag,
        epoch=args.epoch,
        new_size=args.size,
        quantized=args.quantized,
        inference=True,
        reparameterized=args.reparameterized,
        pts=args.pts,
        pt2=args.pt2,
        st=args.st,
        dtype=model_dtype,
    )
    logger.debug(f"RGB stats: {rgb_stats}")

    if args.ignore_dir_names is True:
        num_classes = len(class_to_idx)
        class_to_idx = fs_ops.class_to_idx_from_paths(args.data_path, hierarchical=args.hierarchical)
        assert len(class_to_idx) == num_classes

    if args.class_mapping is not None:
        num_classes = len(class_to_idx)
        class_mapping = fs_ops.read_json_class_file(args.class_mapping)
        idx_to_class = dict(zip(class_to_idx.values(), class_to_idx.keys()))
        idx_to_class.update(dict(zip(class_mapping.values(), class_mapping.keys())))
        class_to_idx = dict(zip(idx_to_class.values(), idx_to_class.keys()))
        assert len(class_to_idx) == num_classes

    if args.show_class is not None:
        if args.show_class not in class_to_idx:
            logger.warning("Selected 'show class' is not part of the model classes")

    if args.fast_matmul is True or args.amp is True:
        torch.set_float32_matmul_precision("high")

    if args.parallel is True and torch.cuda.device_count() > 1:
        compile_methods = ["embedding"] if args.save_embeddings is True else None
        net = InferenceDataParallel(
            net, output_device="cpu", compile_replicas=args.compile, compile_methods=compile_methods
        )
    elif args.compile is True:
        net = torch.compile(net)
        if args.save_embeddings is True:
            net.embedding = torch.compile(net.embedding)

    if args.size is None:
        args.size = lib.get_size_from_signature(signature)
        logger.debug(f"Using size={args.size}")

    batch_size = args.batch_size
    inference_transform = inference_preset(args.size, rgb_stats, args.center_crop, args.simple_crop)
    if args.wds is True:
        wds_path: str | list[str]
        if args.wds_info is not None:
            (wds_path, dataset_size) = wds_args_from_info(args.wds_info, args.wds_split)
            if args.wds_size is not None:
                dataset_size = args.wds_size
        else:
            (wds_path, dataset_size) = prepare_wds_args(args.data_path[0], args.wds_size, device)

        num_samples = dataset_size
        dataset = make_wds_dataset(
            wds_path,
            dataset_size=dataset_size,
            shuffle=args.shuffle,
            samples_names=True,
            transform=inference_transform,
        )
        inference_loader = make_wds_loader(
            dataset,
            batch_size,
            num_workers=args.num_workers,
            prefetch_factor=args.prefetch_factor,
            collate_fn=None,
            world_size=1,
            pin_memory=False,
            exact=True,
        )

    else:
        dataset = make_image_dataset(
            args.data_path, class_to_idx, transforms=inference_transform, hierarchical=args.hierarchical
        )
        num_samples = len(dataset)
        inference_loader = DataLoader(
            dataset,
            batch_size=batch_size,
            shuffle=args.shuffle,
            num_workers=args.num_workers,
            prefetch_factor=args.prefetch_factor,
        )

    show_flag = (
        args.show is True
        or args.show_top_below is not None
        or args.show_target_below is not None
        or args.show_mistakes is True
        or args.show_out_of_k is True
        or args.show_class is not None
    )

    def batch_callback(file_paths: list[str], out: npt.NDArray[np.float32], batch_labels: list[int]) -> None:
        # Show flags
        if show_flag is True:
            for img_path, prob, label in zip(file_paths, out, batch_labels):
                handle_show_flags(args, img_path, prob, label, class_to_idx)

    # Sort out output file names
    epoch_str = ""
    if args.epoch is not None:
        epoch_str = f"_e{args.epoch}"

    base_output_path = (
        f"{network_name}_{len(class_to_idx)}{epoch_str}_{args.size[0]}px_crop{args.center_crop}_{num_samples}"
    )
    if args.simple_crop is True:
        base_output_path = f"{base_output_path}_sc"
    if args.tta is True:
        base_output_path = f"{base_output_path}_tta"
    if args.model_dtype != "float32":
        base_output_path = f"{base_output_path}_{args.model_dtype}"
    if args.prefix is not None:
        base_output_path = f"{args.prefix}_{base_output_path}"
    if args.suffix is not None:
        base_output_path = f"{base_output_path}_{args.suffix}"

    if args.save_sparse_results is True:
        results_file_suffix = "_sparse.csv"
    else:
        results_file_suffix = ".csv"

    results_path = f"{base_output_path}{results_file_suffix}"
    embeddings_path = settings.RESULTS_DIR.joinpath(f"{base_output_path}_embeddings.{args.output_format}")
    logits_path = settings.RESULTS_DIR.joinpath(f"{base_output_path}_logits.{args.output_format}")
    output_path = settings.RESULTS_DIR.joinpath(f"{base_output_path}_output.{args.output_format}")
    label_names: Optional[list[str]] = list(class_to_idx.keys())
    if args.save_logits or args.save_output:
        if label_names is not None and len(label_names) == 0:
            logger.warning("No class labels found, using numeric indices instead")
            # In case that a 'pts' or 'pt2' type models were loaded, we don't know the output size
            # so we’ll fill this lazily on the first batch
            label_names = None

    # Define Parquet writers
    embeddings_writer: pq.ParquetWriter | None = None
    logits_writer: pq.ParquetWriter | None = None
    output_writer: pq.ParquetWriter | None = None

    # Inference
    tic = time.time()
    infer_iter = infer_dataloader_iter(
        device,
        net,
        inference_loader,
        args.save_embeddings,
        args.tta,
        return_logits=args.save_logits,
        model_dtype=model_dtype,
        amp=args.amp,
        amp_dtype=amp_dtype,
        num_samples=num_samples,
        batch_callback=batch_callback,
        chunk_size=args.chunk_size,
        **args.forward_kwargs,
    )
    append = False  # Append mode for outputs, only False for the first batch
    summary_list = []
    with torch.inference_mode():
        for sample_paths, outs, labels, embedding_list in infer_iter:
            if label_names is None:
                label_names = [str(i) for i in range(outs.shape[1])]
                logger.warning(f"Falling back to {len(label_names)} numeric class names")

            # Save embeddings
            if args.save_embeddings is True:
                if args.output_format == "parquet":
                    if embeddings_writer is None:
                        embeddings_writer = _init_array_parquet_writer(
                            ["sample"], "embedding", outs.shape[1], embeddings_path
                        )

                    save_embeddings_parquet(embeddings_writer, sample_paths, embedding_list)
                else:
                    save_embeddings(embeddings_path, sample_paths, embedding_list, append=append)

            if args.save_logits is True:
                if args.output_format == "parquet":
                    if logits_writer is None:
                        logits_writer = _init_parquet_writer(["sample"], label_names, logits_path)

                    save_logits_parquet(logits_writer, sample_paths, outs)
                else:
                    save_logits(logits_path, sample_paths, label_names, outs, append=append)
            elif args.save_output is True:
                if args.output_format == "parquet":
                    if output_writer is None:
                        output_writer = _init_parquet_writer(["sample", "prediction"], label_names, output_path)

                    save_output_parquet(output_writer, sample_paths, label_names, outs)
                else:
                    save_output(output_path, sample_paths, label_names, outs, append=append)

            if len(class_to_idx) > 0:
                # Handle results (Results / SparseResults)
                results: Results | SparseResults
                if args.save_sparse_results is True:
                    results = SparseResults(sample_paths, labels, label_names, output=outs)
                else:
                    results = Results(sample_paths, labels, label_names, output=outs)

                if results.missing_all_labels is False:
                    if args.save_results is True or args.save_sparse_results is True:
                        results.save(results_path, append=append)
                    if args.chunk_size is None:
                        results.log_short_report()

                else:
                    logger.warning("No labeled samples found")

                # Summary
                if args.summary is True:
                    summary_list.append(results.prediction_names.value_counts())

            append = True

    if embeddings_writer is not None:
        embeddings_writer.close()
    if logits_writer is not None:
        logits_writer.close()
    if output_writer is not None:
        output_writer.close()

    toc = time.time()
    rate = num_samples / (toc - tic)
    logger.info(f"{lib.format_duration(toc-tic)} to classify {num_samples:,} samples ({rate:.2f} samples/sec)")

    #  Print summary
    if args.summary is True:
        summary_df = pl.concat(summary_list).group_by("prediction_names").agg(pl.col("count").sum())
        summary_df = summary_df.sort(by="count", descending=True)
        indent_size = summary_df["prediction_names"].str.len_chars().max() + 2  # type: ignore[operator]
        for specie_name, count in summary_df.iter_rows():
            logger.info(f"{specie_name:<{indent_size}} {count}")


def get_args_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        description="Run prediction on directories and/or files",
        epilog=(
            "Usage example:\n"
            "python -m birder.scripts.predict --network resnet_v2_50 --pts --gpu --save-output data/Unknown\n"
            "python predict.py -n fastvit_t8 -t il-common_reparameterized -e 0 --batch-size 256 --reparameterized "
            "--gpu --save-results data/validation_il-common_packed\n"
            "python predict.py --network inception_resnet_v2 -e 100 --gpu --show-out-of-k data/validation\n"
            "python predict.py --network inception_v3 --gpu --batch-size 256 --save-results data/validation/*crane\n"
            "python predict.py -n efficientnet_v2_m -e 0 --gpu --save-embeddings data/testing\n"
            "python predict.py -n efficientnet_v1_b4 -e 300 --gpu --save-embeddings "
            "data/*/Alpine\\ swift --suffix alpine_swift\n"
            "python predict.py -n mobilevit_v2_1_5 -t intermediate -e 80 --gpu --save-results "
            "--wds data/validation_packed\n"
            "python -m birder.scripts.predict -n rope_i_vit_l14_pn_ap_c1 -t pe-core --gpu --parallel --batch-size 256 "
            "--chunk-size 50000 --amp --amp-dtype bfloat16 --compile --save-embeddings "
            "--suffix raw_data data/raw_data\n"
            "python predict.py -n convnext_v2_tiny -t intermediate -e 70 --gpu --gpu-id 1 --compile --fast-matmul "
            "--show-class Unknown data/raw_data\n"
            "python -m birder.scripts.predict -n rope_vit_reg4_b14 --gpu --amp --ignore-dir-names data/imagenet-v2\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    parser.add_argument("-n", "--network", type=str, help="the neural network to use (i.e. resnet_v2)")
    parser.add_argument(
        "--model-config",
        action=cli.FlexibleDictAction,
        help=(
            "override the model default configuration, accepts key-value pairs or JSON "
            "('drop_path_rate=0.2' or '{\"units\": [3, 24, 36, 3], \"dropout\": 0.2}'"
        ),
    )
    parser.add_argument(
        "--forward-kwargs",
        action=cli.FlexibleDictAction,
        help=(
            "additional model forward/embedding keyword args, accepts key-value pairs or JSON "
            "('patch_size=12' or '{\"patch_size\": 20}'"
        ),
    )
    parser.add_argument("-e", "--epoch", type=int, metavar="N", help="model checkpoint to load")
    parser.add_argument("--quantized", default=False, action="store_true", help="load quantized model")
    parser.add_argument("-t", "--tag", type=str, help="model tag (from the training phase)")
    parser.add_argument(
        "-r", "--reparameterized", default=False, action="store_true", help="load reparameterized model"
    )
    parser.add_argument("--pts", default=False, action="store_true", help="load torchscript network")
    parser.add_argument("--pt2", default=False, action="store_true", help="load standardized model")
    parser.add_argument("--st", "--safetensors", default=False, action="store_true", help="load Safetensors weights")
    parser.add_argument("--compile", default=False, action="store_true", help="enable compilation")
    parser.add_argument(
        "--model-dtype",
        type=str,
        choices=["float32", "float16", "bfloat16"],
        default="float32",
        help="model dtype to use",
    )
    parser.add_argument(
        "--amp", default=False, action="store_true", help="use torch.amp.autocast for mixed precision inference"
    )
    parser.add_argument(
        "--amp-dtype",
        type=str,
        choices=["float16", "bfloat16"],
        help="whether to use float16 or bfloat16 for mixed precision",
    )
    parser.add_argument(
        "--fast-matmul", default=False, action="store_true", help="use fast matrix multiplication (affects precision)"
    )
    parser.add_argument("--tta", default=False, action="store_true", help="test time augmentation (oversampling)")
    parser.add_argument(
        "--size", type=int, nargs="+", metavar=("H", "W"), help="image size for inference (defaults to model signature)"
    )
    parser.add_argument("--batch-size", type=int, default=32, metavar="N", help="the batch size")
    parser.add_argument("-j", "--num-workers", type=int, default=8, metavar="N", help="number of preprocessing workers")
    parser.add_argument(
        "--prefetch-factor", type=int, metavar="N", help="number of batches loaded in advance by each worker"
    )
    parser.add_argument(
        "--chunk-size", type=int, metavar="N", help="process in chunks of N samples to reduce memory usage"
    )
    parser.add_argument("--center-crop", type=float, default=1.0, help="center crop ratio to use during inference")
    parser.add_argument(
        "--simple-crop",
        default=False,
        action="store_true",
        help="use a simple crop that preserves aspect ratio but may trim parts of the image",
    )
    parser.add_argument("--show", default=False, action="store_true", help="show image predictions")
    parser.add_argument("--show-top-below", type=float, help="show when top prediction is below given threshold")
    parser.add_argument("--show-target-below", type=float, help="show when target prediction is below given threshold")
    parser.add_argument("--show-mistakes", default=False, action="store_true", help="show only mis-classified images")
    parser.add_argument("--show-out-of-k", default=False, action="store_true", help="show images not in the top-k")
    parser.add_argument("--show-class", type=str, help="show specific class predictions")
    parser.add_argument("--shuffle", default=False, action="store_true", help="predict samples in random order")
    parser.add_argument("--summary", default=False, action="store_true", help="log prediction summary")
    parser.add_argument("--save-results", default=False, action="store_true", help="save results object")
    parser.add_argument(
        "--save-sparse-results",
        default=False,
        action="store_true",
        help="save results object in memory-efficient sparse format (only top-k probabilities)",
    )
    parser.add_argument("--save-output", default=False, action="store_true", help="save raw output")
    parser.add_argument("--save-logits", default=False, action="store_true", help="save raw model logits")
    parser.add_argument("--save-embeddings", default=False, action="store_true", help="save embedding layer outputs")
    parser.add_argument(
        "--output-format",
        type=str,
        choices=["csv", "parquet"],
        default="csv",
        help="output format for raw data (embeddings, logits, output) - does not affect results files",
    )
    parser.add_argument("--prefix", type=str, help="add prefix to output file")
    parser.add_argument("--suffix", type=str, help="add suffix to output file")
    parser.add_argument("--gpu", default=False, action="store_true", help="use gpu")
    parser.add_argument("--gpu-id", type=int, metavar="ID", help="gpu id to use (ignored in parallel mode)")
    parser.add_argument("--mps", default=False, action="store_true", help="use mps (Metal Performance Shaders) device")
    parser.add_argument("--parallel", default=False, action="store_true", help="use multiple gpus")
    parser.add_argument("--wds", default=False, action="store_true", help="predict a webdataset directory")
    parser.add_argument("--wds-size", type=int, metavar="N", help="size of the wds dataset")
    parser.add_argument("--wds-info", type=str, metavar="FILE", help="wds info file path")
    parser.add_argument("--wds-split", type=str, default="validation", metavar="NAME", help="wds dataset split to load")
    parser.add_argument(
        "--ignore-dir-names",
        default=False,
        action="store_true",
        help="use directory indices (0, 1, 2, ...) as class labels instead of directory names",
    )
    parser.add_argument(
        "--class-mapping", type=str, help="json file with alternative class names (can be partial mapping)"
    )
    parser.add_argument(
        "--hierarchical",
        default=False,
        action="store_true",
        help="use hierarchical directory structure for labels (e.g., 'dir1/subdir2' -> 'dir1_subdir2' label)",
    )
    parser.add_argument("data_path", nargs="*", help="data files path (directories and files)")

    return parser


def validate_args(args: argparse.Namespace) -> None:
    args.size = cli.parse_size(args.size)
    if args.forward_kwargs is None:
        args.forward_kwargs = {}

    if args.network is None:
        raise cli.ValidationError("--network is required")
    if args.center_crop > 1 or args.center_crop <= 0.0:
        raise cli.ValidationError(f"--center-crop must be in range of (0, 1.0], got {args.center_crop}")
    if args.parallel is True and args.gpu is False:
        raise cli.ValidationError("--parallel requires --gpu to be set")
    if args.save_results is True and args.save_sparse_results is True:
        raise cli.ValidationError("--save-results cannot be used with --save-sparse-results")
    if args.save_embeddings is True and args.tta is True:
        raise cli.ValidationError("--save-embeddings cannot be used with --tta")
    if args.amp is True and args.model_dtype != "float32":
        raise cli.ValidationError("--amp can only be used with --model-dtype float32")

    if args.save_logits is True and args.save_results is True:
        raise cli.ValidationError("--save-logits cannot be used with --save-results")
    if args.save_logits is True and args.save_sparse_results is True:
        raise cli.ValidationError("--save-logits cannot be used with --save-sparse-results")
    if args.save_logits is True and args.save_output is True:
        raise cli.ValidationError("--save-logits cannot be used with --save-output")
    if args.save_logits is True and args.summary is True:
        raise cli.ValidationError("--save-logits cannot be used with --summary")

    if args.save_logits is True and (  # pylint: disable=too-many-boolean-expressions
        args.show is True
        or args.show_top_below is not None
        or args.show_target_below is not None
        or args.show_mistakes is True
        or args.show_out_of_k is True
        or args.show_class is not None
    ):
        # Results are not calculated
        raise cli.ValidationError("--save-logits cannot be used with any of the 'show' flags")

    if args.wds is False and len(args.data_path) == 0:
        raise cli.ValidationError("Must provide at least one data source, --data-path or --wds")
    if args.wds is True and len(args.data_path) > 1:
        raise cli.ValidationError(f"--wds can have at most 1 --data-path, got {len(args.data_path)}")
    if args.wds is True and args.hierarchical is True:
        raise cli.ValidationError("--wds cannot be used with --hierarchical")
    if args.wds is True and args.ignore_dir_names is True:
        raise cli.ValidationError("--wds cannot be used with --ignore-dir-names")
    if args.wds is True and (  # pylint: disable=too-many-boolean-expressions
        args.show is True
        or args.show_top_below is not None
        or args.show_target_below is not None
        or args.show_mistakes is True
        or args.show_out_of_k is True
        or args.show_class is not None
    ):
        raise cli.ValidationError("--wds cannot be used with any of the 'show' flags")


def args_from_dict(**kwargs: Any) -> argparse.Namespace:
    parser = get_args_parser()
    parser.set_defaults(**kwargs)
    args = parser.parse_args([])
    validate_args(args)

    return args


def main() -> None:
    parser = get_args_parser()
    args = parser.parse_args()
    validate_args(args)

    if settings.RESULTS_DIR.exists() is False:
        logger.info(f"Creating {settings.RESULTS_DIR} directory...")
        settings.RESULTS_DIR.mkdir(parents=True)

    predict(args)


if __name__ == "__main__":
    logger = logging.getLogger(getattr(__spec__, "name", __name__))
    main()



"""
HorNet, adapted from
https://github.com/raoyongming/HorNet/blob/master/hornet.py

Paper "HorNet: Efficient High-Order Spatial Interactions with Recursive Gated Convolutions",
https://arxiv.org/abs/2207.14284
"""

# Reference license: MIT

from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import StochasticDepth

from birder.layers import LayerScale2d
from birder.model_registry import registry
from birder.net.base import DetectorBackbone


class ChannelsFirstLayerNorm(nn.Module):
    """
    Channels first corresponds to inputs with shape (batch_size, channels, height, width)
    """

    def __init__(self, dim: int, eps: float = 1e-6) -> None:
        super().__init__()
        self.weight = nn.Parameter(torch.ones(dim))
        self.bias = nn.Parameter(torch.zeros(dim))
        self.eps = nn.Buffer(torch.tensor(eps))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        u = x.mean(1, keepdim=True)
        s = (x - u).pow(2).mean(1, keepdim=True)
        x = (x - u) / torch.sqrt(s + self.eps)
        x = self.weight[:, None, None] * x + self.bias[:, None, None]

        return x


class GlobalLocalFilter(nn.Module):
    def __init__(self, dim: int, h: int, w: int) -> None:
        super().__init__()
        self.pre_norm = ChannelsFirstLayerNorm(dim, eps=1e-6)
        self.dw = nn.Conv2d(
            dim // 2, dim // 2, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), groups=dim // 2, bias=False
        )
        self.complex_weight = nn.Parameter(torch.empty(dim // 2, h, w, 2, dtype=torch.float32))
        self.post_norm = ChannelsFirstLayerNorm(dim, eps=1e-6)

        # Weight initialization
        nn.init.trunc_normal_(self.complex_weight, std=0.02)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.pre_norm(x)
        (x1, x2) = torch.chunk(x, 2, dim=1)
        x1 = self.dw(x1)

        x2 = x2.to(torch.float32)
        (B, C, a, b) = x2.size()
        x2 = torch.fft.rfft2(x2, dim=(2, 3), norm="ortho")  # pylint: disable=not-callable

        weight = torch.view_as_complex(self.complex_weight.contiguous())
        x2 = x2 * weight
        x2 = torch.fft.irfft2(x2, s=(a, b), dim=(2, 3), norm="ortho")  # pylint: disable=not-callable

        x = torch.concat([x1.unsqueeze(2), x2.unsqueeze(2)], dim=2).reshape(B, 2 * C, a, b)
        x = self.post_norm(x)

        return x


# pylint: disable=invalid-name
class gnConv(nn.Module):
    def __init__(self, dim: int, order: int, gf_layer: bool, h: int, w: int, scale: float) -> None:
        super().__init__()
        self.order = order
        self.dims = [dim // 2**i for i in range(order)]
        self.dims.reverse()
        self.proj_in = nn.Conv2d(dim, 2 * dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))

        if gf_layer is False:
            self.dwconv = nn.Conv2d(
                sum(self.dims), sum(self.dims), kernel_size=(7, 7), stride=(1, 1), padding=(3, 3), groups=sum(self.dims)
            )
        else:
            self.dwconv = GlobalLocalFilter(sum(self.dims), h=h, w=w)

        self.proj_out = nn.Conv2d(dim, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))

        self.pws = nn.ModuleList(
            [
                nn.Conv2d(self.dims[i], self.dims[i + 1], kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
                for i in range(order - 1)
            ]
        )

        self.scale = scale

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        fused_x = self.proj_in(x)
        (pwa, abc) = torch.split(fused_x, (self.dims[0], sum(self.dims)), dim=1)
        dw_abc = self.dwconv(abc) * self.scale
        dw_list = torch.split(dw_abc, self.dims, dim=1)
        x = pwa * dw_list[0]

        for i, mod in enumerate(self.pws):
            x = mod(x) * dw_list[i + 1]

        x = self.proj_out(x)

        return x


class HorBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        layer_scale_init_value: Optional[float],
        order: int,
        gf_layer: bool,
        h: int,
        w: int,
        scale: float,
        drop_path: float,
    ) -> None:
        super().__init__()

        self.norm1 = ChannelsFirstLayerNorm(dim, eps=1e-6)
        self.gn_conv = gnConv(dim, order=order, gf_layer=gf_layer, h=h, w=w, scale=scale)
        self.norm2 = nn.LayerNorm(dim, eps=1e-6)
        self.pw_conv1 = nn.Linear(dim, 4 * dim)
        self.act = nn.GELU()
        self.pw_conv2 = nn.Linear(4 * dim, dim)

        if layer_scale_init_value is not None:
            self.layer_scale_1 = LayerScale2d(dim, layer_scale_init_value)
            self.layer_scale_2 = LayerScale2d(dim, layer_scale_init_value)
        else:
            self.layer_scale_1 = nn.Identity()
            self.layer_scale_2 = nn.Identity()

        self.drop_path = StochasticDepth(drop_path, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        res = x
        x = self.norm1(x)
        x = self.gn_conv(x)
        x = self.layer_scale_1(x)
        x = res + self.drop_path(x)

        res = x
        x = x.permute(0, 2, 3, 1)  # (N, C, H, W) -> (N, H, W, C)
        x = self.norm2(x)
        x = self.pw_conv1(x)
        x = self.act(x)
        x = self.pw_conv2(x)
        x = x.permute(0, 3, 1, 2)  # (N, H, W, C) -> (N, C, H, W)

        x = self.layer_scale_2(x)
        x = res + self.drop_path(x)

        return x


class HorStage(nn.Module):
    def __init__(
        self,
        dim: int,
        out_dim: int,
        layer_scale_init_value: Optional[float],
        order: int,
        gf_layer: bool,
        h: int,
        w: int,
        scale: float,
        drop_path: list[float],
        depth: int,
        downsample: bool,
    ) -> None:
        super().__init__()
        if downsample is True:
            self.downsample = nn.Sequential(
                ChannelsFirstLayerNorm(dim, eps=1e-6),
                nn.Conv2d(dim, out_dim, kernel_size=(2, 2), stride=(2, 2), padding=(0, 0)),
            )
        else:
            assert dim == out_dim
            self.downsample = nn.Identity()

        layers = []
        for i in range(depth):
            layers.append(
                HorBlock(
                    out_dim,
                    layer_scale_init_value=layer_scale_init_value,
                    order=order,
                    gf_layer=gf_layer,
                    h=h,
                    w=w,
                    scale=scale,
                    drop_path=drop_path[i],
                )
            )

        self.blocks = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.downsample(x)
        x = self.blocks(x)

        return x


class HorNet(DetectorBackbone):
    square_only = True

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        s = 1.0 / 3.0
        layer_scale_init_value = 1e-6
        gn_conv_order = [2, 3, 4, 5]
        depths: list[int] = self.config["depths"]
        base_dim: int = self.config["base_dim"]
        gf_layer: list[bool] = self.config["gf_layer"]
        drop_path_rate: float = self.config["drop_path_rate"]

        dpr = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(depths)).split(depths)]
        num_stages = len(depths)
        dims = [base_dim, base_dim * 2, base_dim * 4, base_dim * 8]
        gn_conv_h = [self.size[0] // 4, self.size[0] // 8, self.size[0] // 16, self.size[0] // 32]
        gn_conv_w = [(h // 2) + 1 for h in gn_conv_h]

        self.stem = nn.Sequential(
            nn.Conv2d(self.input_channels, dims[0], kernel_size=(4, 4), stride=(4, 4), padding=(0, 0)),
            ChannelsFirstLayerNorm(dims[0], eps=1e-6),
        )

        prev_dim = dims[0]
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i in range(num_stages):
            stages[f"stage{i+1}"] = HorStage(
                prev_dim,
                dims[i],
                layer_scale_init_value=layer_scale_init_value,
                order=gn_conv_order[i],
                gf_layer=gf_layer[i],
                h=gn_conv_h[i],
                w=gn_conv_w[i],
                scale=s,
                drop_path=dpr[i],
                depth=depths[i],
                downsample=i > 0,
            )
            return_channels.append(dims[i])
            prev_dim = dims[i]

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
            nn.LayerNorm(dims[-1], eps=1e-6),
        )
        self.return_channels = return_channels
        self.embedding_size = dims[-1]
        self.classifier = self.create_classifier()

        # Weight initialization
        for m in self.modules():
            if isinstance(m, (nn.Conv2d, nn.Linear)):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:
        assert dynamic_size is False, "Dynamic size not supported for this network"

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        super().adjust_size(new_size)

        gn_conv_h = [new_size[0] // 4, new_size[0] // 8, new_size[0] // 16, new_size[0] // 32]
        gn_conv_w = [(h // 2) + 1 for h in gn_conv_h]
        for i, module in enumerate(self.body.children()):
            if isinstance(module, HorStage):
                for m in module.modules():
                    if isinstance(m, HorBlock):
                        if isinstance(m.gn_conv.dwconv, GlobalLocalFilter):
                            weight = m.gn_conv.dwconv.complex_weight
                            weight = F.interpolate(
                                weight.permute(3, 0, 1, 2),
                                size=(gn_conv_h[i], gn_conv_w[i]),
                                mode="bilinear",
                                align_corners=True,
                            ).permute(1, 2, 3, 0)
                            m.gn_conv.dwconv.complex_weight = nn.Parameter(weight)


registry.register_model_config(
    "hornet_tiny_7x7",
    HorNet,
    config={
        "depths": [2, 3, 18, 2],
        "base_dim": 64,
        "gf_layer": [False, False, False, False],
        "drop_path_rate": 0.2,
    },
)
registry.register_model_config(
    "hornet_tiny_gf",
    HorNet,
    config={
        "depths": [2, 3, 18, 2],
        "base_dim": 64,
        "gf_layer": [False, False, True, True],
        "drop_path_rate": 0.2,
    },
)
registry.register_model_config(
    "hornet_small_7x7",
    HorNet,
    config={
        "depths": [2, 3, 18, 2],
        "base_dim": 96,
        "gf_layer": [False, False, False, False],
        "drop_path_rate": 0.4,
    },
)
registry.register_model_config(
    "hornet_small_gf",
    HorNet,
    config={
        "depths": [2, 3, 18, 2],
        "base_dim": 96,
        "gf_layer": [False, False, True, True],
        "drop_path_rate": 0.4,
    },
)
registry.register_model_config(
    "hornet_base_7x7",
    HorNet,
    config={
        "depths": [2, 3, 18, 2],
        "base_dim": 128,
        "gf_layer": [False, False, False, False],
        "drop_path_rate": 0.5,
    },
)
registry.register_model_config(
    "hornet_base_gf",
    HorNet,
    config={
        "depths": [2, 3, 18, 2],
        "base_dim": 128,
        "gf_layer": [False, False, True, True],
        "drop_path_rate": 0.5,
    },
)
registry.register_model_config(
    "hornet_large_7x7",
    HorNet,
    config={
        "depths": [2, 3, 18, 2],
        "base_dim": 192,
        "gf_layer": [False, False, False, False],
        "drop_path_rate": 0.4,
    },
)
registry.register_model_config(
    "hornet_large_gf",
    HorNet,
    config={
        "depths": [2, 3, 18, 2],
        "base_dim": 192,
        "gf_layer": [False, False, True, True],
        "drop_path_rate": 0.4,
    },
)

registry.register_weights(
    "hornet_tiny_7x7_danube-delta",
    {
        "url": "https://huggingface.co/birder-project/hornet_tiny_7x7_danube-delta/resolve/main",
        "description": "HorNet tiny 7x7 model trained on the danube-delta dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 84.5,
                "sha256": "d0555e97c5f353d5659fe518d43c3c7921e90987826b95dff024032ec06fb79a",
            }
        },
        "net": {"network": "hornet_tiny_7x7", "tag": "danube-delta"},
    },
)



"""
Adapted from https://github.com/jacobgil/vit-explain/blob/main/vit_rollout.py
"""

# Reference license: MIT

from collections.abc import Callable
from pathlib import Path
from typing import Literal
from typing import Optional

import numpy as np
import torch
from PIL import Image
from torch import nn

from birder.introspection.base import InterpretabilityResult
from birder.introspection.base import Interpreter
from birder.introspection.base import show_mask_on_image
from birder.net.vit import Encoder


def rollout(
    attentions: list[torch.Tensor],
    discard_ratio: float,
    head_fusion: Literal["mean", "max", "min"],
    num_special_tokens: int,
) -> torch.Tensor:
    result = torch.eye(attentions[0].size(-1))
    with torch.no_grad():
        for attention in attentions:
            if head_fusion == "mean":
                attention_heads_fused = attention.mean(axis=1)
            elif head_fusion == "max":
                attention_heads_fused = attention.max(axis=1)[0]
            elif head_fusion == "min":
                attention_heads_fused = attention.min(axis=1)[0]
            else:
                raise ValueError("Attention head fusion type Not supported")

            # Drop the lowest attentions, but don't drop the class token
            flat = attention_heads_fused.view(attention_heads_fused.size(0), -1)
            (_, indices) = flat.topk(int(flat.size(-1) * discard_ratio), -1, False)
            indices = indices[indices != 0]
            flat[0, indices] = 0

            eye = torch.eye(attention_heads_fused.size(-1))
            a = (attention_heads_fused + 1.0 * eye) / 2
            a = a / a.sum(dim=-1)

            result = torch.matmul(a, result)

    # Look at the total attention between the class token and the image patches
    mask = result[0, 0, num_special_tokens:]

    width = int(mask.size(-1) ** 0.5)
    mask = mask.reshape(width, width)
    mask = mask / torch.max(mask)

    return mask


class AttentionRollout:
    def __init__(self, net: torch.nn.Module, attention_layer_name: str) -> None:
        assert hasattr(net, "encoder") is True and isinstance(net.encoder, Encoder)
        net.encoder.set_need_attn()
        self.net = net
        for name, module in self.net.named_modules():
            if name.endswith(attention_layer_name) is True:
                module.register_forward_hook(self.get_attention)

        self.attentions: list[torch.Tensor] = []

    def get_attention(
        self, _module: torch.nn.Module, _inputs: tuple[torch.Tensor, ...], outputs: tuple[torch.Tensor, ...]
    ) -> None:
        self.attentions.append(outputs[1].cpu())

    def __call__(
        self, x: torch.Tensor, discard_ratio: float, head_fusion: Literal["mean", "max", "min"]
    ) -> torch.Tensor:
        self.attentions = []
        with torch.inference_mode():
            self.net(x)

        return rollout(self.attentions, discard_ratio, head_fusion, self.net.num_special_tokens)


class AttentionRolloutInterpreter(Interpreter):
    def __init__(
        self,
        model: nn.Module,
        device: torch.device,
        transform: Callable[..., torch.Tensor],
        attention_layer_name: str,
        discard_ratio: float,
        head_fusion: Literal["mean", "max", "min"],
    ) -> None:
        super().__init__(model, device, transform)
        self.attention_rollout = AttentionRollout(model, attention_layer_name)
        self.discard_ratio = discard_ratio
        self.head_fusion = head_fusion

    def interpret(self, image: str | Path | Image.Image, target_class: Optional[int] = None) -> InterpretabilityResult:
        (input_tensor, rgb_img) = self._preprocess_image(image)

        attention_map = self.attention_rollout(
            input_tensor, discard_ratio=self.discard_ratio, head_fusion=self.head_fusion
        )

        # Resize attention map to match image size
        attention_img = Image.fromarray(attention_map.numpy())
        attention_img = attention_img.resize(rgb_img.shape[:2])
        attention_arr = np.array(attention_img)
        visualization = show_mask_on_image(rgb_img, attention_arr, image_weight=0.4)

        return InterpretabilityResult(rgb_img, visualization, raw_output=attention_arr)



import argparse
import json
import logging
import multiprocessing
import os
import signal
import time
from collections.abc import Callable
from io import BytesIO
from pathlib import Path
from typing import Any
from typing import Optional

import torch
import webdataset as wds
from PIL import Image
from torch.utils.data import ConcatDataset
from torchvision.datasets import ImageFolder
from tqdm import tqdm

from birder.common import cli
from birder.common import fs_ops
from birder.common.lib import format_duration
from birder.conf import settings

logger = logging.getLogger(__name__)

# Few datasets like Objects365 have some very big files
Image.MAX_IMAGE_PIXELS = int(2048 * 2048 * 1024 // 4 // 3)
MAX_SIZE = 16_000


class CustomImageFolder(ImageFolder):
    def __init__(
        self,
        root: str,
        *,
        class_to_idx: dict[str, int],
    ) -> None:
        self._class_to_idx = class_to_idx
        super().__init__(root, loader=str, allow_empty=True)

    def find_classes(self, _directory: str) -> tuple[list[str], dict[str, int]]:
        classes = list(self._class_to_idx.keys())
        return (classes, self._class_to_idx)


def _get_class_to_idx(paths: list[str]) -> dict[str, int]:
    class_list: list[str] = []
    for path in paths:
        dataset = ImageFolder(path)
        class_list.extend(list(dataset.class_to_idx.keys()))

    class_list = sorted(list(set(class_list)))
    class_to_idx = {k: v for v, k in enumerate(class_list)}

    return class_to_idx


def _save_classes(pack_path: Path, class_to_idx: dict[str, int]) -> None:
    class_list_path = pack_path.joinpath("classes.txt")
    doc = "\n".join(list(class_to_idx.keys()))

    logger.info(f"Saving class list at {class_list_path}")
    with open(class_list_path, "w", encoding="utf-8") as handle:
        handle.write(doc)


def _encode_image(path: str, file_format: str, size: Optional[int] = None) -> bytes:
    image: Image.Image = Image.open(path)
    if size is not None and size < min(image.size):
        if image.size[0] > image.size[1]:
            ratio = image.size[1] / size
        else:
            ratio = image.size[0] / size

        width = round(image.size[0] / ratio)
        height = round(image.size[1] / ratio)
        if max(width, height) > MAX_SIZE:
            if width > height:
                ratio = width / MAX_SIZE
            else:
                ratio = height / MAX_SIZE

            width = round(width / ratio)
            height = round(height / ratio)

        image = image.resize((width, height), Image.Resampling.BICUBIC)

    elif max(image.size) > MAX_SIZE:
        if image.size[0] > image.size[1]:
            ratio = image.size[0] / MAX_SIZE
        else:
            ratio = image.size[1] / MAX_SIZE

        width = round(image.size[0] / ratio)
        height = round(image.size[1] / ratio)
        image = image.resize((width, height), Image.Resampling.BICUBIC)

    sample_buffer = BytesIO()
    image.save(sample_buffer, format=file_format, quality=85)
    return sample_buffer.getvalue()


def read_worker(q_in: Any, q_out: Any, error_event: Any, size: Optional[int], file_format: str) -> None:
    while True:
        deq: Optional[tuple[int, str, int]] = q_in.get()
        if deq is None:
            break

        try:
            (idx, path, target) = deq
            if size is None:
                suffix = Path(path).suffix[1:]
                if file_format != suffix:
                    sample = _encode_image(path, file_format)
                else:
                    with open(path, "rb") as stream:
                        sample = stream.read()

            else:
                sample = _encode_image(path, file_format, size)

        except Exception:
            error_event.set()
            raise

        if error_event.is_set() is True:
            break

        q_out.put((idx, sample, file_format, target), block=True, timeout=None)


def wds_write_worker(
    q_out: Any, error_event: Any, pack_path: Path, total: int, args: argparse.Namespace, _: dict[int, str]
) -> None:
    try:
        info_path = pack_path.joinpath("_info.json")
        if args.append is True:
            info = fs_ops.read_wds_info(info_path)
            if args.split in info["splits"]:
                raise ValueError(f"split {args.split} already exist")

        else:
            info = None
    except Exception:
        error_event.set()
        # Re-raise to terminate this process
        raise

    filenames: list[str] = []
    shard_lengths: list[int] = []
    path_pattern = str(pack_path.joinpath(f"{args.suffix}-{args.split}-%06d.tar"))
    sink = wds.ShardWriter(path_pattern, maxsize=args.max_size, verbose=0)

    def wds_info(fname: str) -> None:
        filenames.append(Path(fname).name)
        shard_lengths.append(sink.count)

    sink.post = wds_info

    count = 0
    buf = {}
    more = True
    with tqdm(total=total, initial=0, unit="images", unit_scale=True, leave=False) as progress:
        while more:
            deq: Optional[tuple[int, bytes, str, int]] = q_out.get()
            if deq is not None:
                (idx, sample, suffix, target) = deq
                buf[idx] = (sample, suffix, target)

            else:
                more = False

            # Ensures ordered write
            while count in buf:
                (sample, suffix, target) = buf[count]
                del buf[count]

                if args.no_cls is True:
                    cls = {}
                else:
                    cls = {"cls": target}

                sink.write(
                    {
                        "__key__": f"sample{count:09d}",
                        suffix: sample,
                        **cls,
                    }
                )

                count += 1

                # Update progress bar
                progress.update(n=1)

    sink.close()

    split_info = {
        "name": args.split,
        "filenames": filenames,
        "shard_lengths": shard_lengths,
        "num_samples": sum(shard_lengths),
    }

    if info is None:
        info = {
            "name": args.suffix,
            "splits": {args.split: split_info},
        }
    else:
        info["splits"][args.split] = split_info

    with open(pack_path.joinpath("_info.json"), "w", encoding="utf-8") as handle:
        logger.debug("Saving _info.json")
        json.dump(info, handle, indent=2)


def directory_write_worker(
    q_out: Any, _error_event: Any, pack_path: Path, total: int, _: argparse.Namespace, idx_to_class: dict[int, str]
) -> None:
    count = 0
    buf = {}
    more = True
    with tqdm(total=total, initial=0, unit="images", unit_scale=True, leave=False) as progress:
        while more:
            deq: Optional[tuple[int, bytes, str, int]] = q_out.get()
            if deq is not None:
                (idx, sample, suffix, target) = deq
                buf[idx] = (sample, suffix, target)

            else:
                more = False

            # Ensures ordered write
            while count in buf:
                (sample, suffix, target) = buf[count]
                del buf[count]
                with open(pack_path.joinpath(idx_to_class[target]).joinpath(f"{count:06d}.{suffix}"), "wb") as handle:
                    handle.write(sample)

                count += 1

                # Update progress bar
                progress.update(n=1)


# pylint: disable=too-many-locals,too-many-branches
def pack(args: argparse.Namespace, pack_path: Path) -> None:
    if args.sampling_file is not None:
        with open(args.sampling_file, "r", encoding="utf-8") as handle:
            sampling_lines = handle.readlines()

        data_paths = []
        for line in sampling_lines:
            if len(line.strip()) == 0 or line.strip().startswith("#") is True:
                continue

            (data_path, r) = line.split()
            data_path = os.path.expanduser(data_path)
            repeats = int(r)
            for _ in range(repeats):
                data_paths.append(data_path)

    else:
        data_paths = args.data_path

    if args.no_cls is False:
        if args.class_file is not None:
            class_to_idx = fs_ops.read_class_file(args.class_file)
        elif args.append is True:
            class_to_idx = fs_ops.read_class_file(pack_path.joinpath("classes.txt"))
        else:
            class_to_idx = _get_class_to_idx(data_paths)

        _save_classes(pack_path, class_to_idx)
        idx_to_class = dict(zip(class_to_idx.values(), class_to_idx.keys()))

        datasets = []
        for path in data_paths:
            datasets.append(CustomImageFolder(path, class_to_idx=class_to_idx))

        dataset = ConcatDataset(datasets)

    else:
        idx_to_class = {}
        dataset = fs_ops.collect_samples_from_paths(data_paths, class_to_idx={})

    if args.shuffle is True:
        indices = torch.randperm(len(dataset)).tolist()
    else:
        indices = list(range(len(dataset)))

    if args.jobs == -1:
        args.jobs = multiprocessing.cpu_count()

    logger.info(f"Packing {len(dataset):,} samples")
    logger.info(f"Running {args.jobs} read processes and 1 write process")

    q_in = []  # type: ignore
    for _ in range(args.jobs):
        q_in.append(multiprocessing.Queue(1024))

    q_out = multiprocessing.Queue(1024)  # type: ignore
    error_event = multiprocessing.Event()

    read_processes: list[multiprocessing.Process] = []
    for idx in range(args.jobs):
        read_processes.append(
            multiprocessing.Process(
                target=read_worker, args=(q_in[idx], q_out, error_event, args.size, args.format), daemon=True
            )
        )

    for p in read_processes:
        p.start()

    if args.type == "wds":
        target_writer: Callable[..., None] = wds_write_worker
    elif args.type == "directory":
        target_writer = directory_write_worker
        for c in class_to_idx.keys():
            pack_path.joinpath(c).mkdir()
    else:
        raise ValueError("Unknown pack type")

    write_process = multiprocessing.Process(
        target=target_writer, args=(q_out, error_event, pack_path, len(dataset), args, idx_to_class), daemon=True
    )
    write_process.start()

    def signal_handler(signum, _frame) -> None:  # type: ignore
        logger.info(f"Received signal: {signum} at {multiprocessing.current_process().name}, aborting...")
        error_event.set()
        raise SystemExit(1)

    signal.signal(signal.SIGINT, signal_handler)

    tic = time.time()
    for idx, sample_idx in enumerate(indices):
        if idx % 1000 == 0:
            if error_event.is_set() is True:
                raise RuntimeError()

        (path, target) = dataset[sample_idx]
        q_in[idx % len(q_in)].put((idx, path, target), block=True, timeout=None)

    for q in q_in:
        q.put(None, block=True, timeout=None)

    for p in read_processes:
        while True:
            p.join(timeout=2)
            if p.is_alive() is False:
                break

            if error_event.is_set() is True:
                raise RuntimeError()

    q_out.put(None, block=True, timeout=None)
    write_process.join()

    if error_event.is_set() is True:
        raise RuntimeError()

    if args.type == "wds":
        (wds_path, num_shards) = fs_ops.wds_braces_from_path(pack_path, prefix=f"{args.suffix}-{args.split}")
        logger.info(f"Packed {len(dataset):,} samples into {num_shards} shards at {wds_path}")
    elif args.type == "directory":
        logger.info(f"Packed {len(dataset):,} samples")

    toc = time.time()
    rate = len(dataset) / (toc - tic)
    logger.info(f"{format_duration(toc-tic)} to pack {len(dataset):,} samples ({rate:.2f} samples/sec)")


def set_parser(subparsers: Any) -> None:
    subparser = subparsers.add_parser(
        "pack",
        allow_abbrev=False,
        help="pack image dataset",
        description="pack image dataset",
        epilog=(
            "Usage examples:\n"
            "python -m birder.tools pack --size 512 data/training\n"
            "python -m birder.tools pack -j 4 --shuffle --max-size 80 --target-path data/cub_200_2011 "
            "--suffix cub_200_2011 data/CUB_200_2011/training\n"
            "python -m birder.tools pack -j 4 --max-size 80 --target-path data/cub_200_2011 "
            "--suffix cub_200_2011 --split validation --append data/CUB_200_2011/validation\n"
            "python -m birder.tools pack --type directory -j 8 --suffix il-common_packed --size 448 "
            "--format jpeg --class-file data/il-common_classes.txt data/training\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparser.add_argument("--type", type=str, choices=["wds", "directory"], default="wds", help="pack type")
    subparser.add_argument("--target-path", type=str, help="where to write the packed dataset")
    subparser.add_argument("--max-size", type=int, default=400, help="maximum size of each shard in MB")
    subparser.add_argument(
        "-j",
        "--jobs",
        type=int,
        default=1,
        help="performs calculation on multiple cores, set -1 to run on all cores",
    )
    subparser.add_argument("--shuffle", default=False, action="store_true", help="shuffle the dataset during packing")
    subparser.add_argument("--size", type=int, help="resize image short dimension to size if bigger")
    subparser.add_argument("--format", type=str, choices=["webp", "png", "jpeg"], default="webp", help="file format")
    subparser.add_argument("--class-file", type=str, help="class list file")
    subparser.add_argument("--no-cls", default=False, action="store_true", help="pack without class information")
    subparser.add_argument("--suffix", type=str, default=settings.PACK_PATH_SUFFIX, help="directory suffix")
    subparser.add_argument("--split", type=str, default="training", help="dataset split used for _info.json")
    subparser.add_argument("--append", default=False, action="store_true", help="add split to existing wds")
    subparser.add_argument(
        "--sampling-file",
        type=str,
        help="file containing dataset paths and their sampling ratios (overrides data_path argument)",
    )
    subparser.add_argument("data_path", nargs="*", help="image directories")
    subparser.set_defaults(func=main)


def main(args: argparse.Namespace) -> None:
    if args.append is True and args.type != "wds":
        raise cli.ValidationError("--append requires --type wds to be set")
    if args.no_cls is True and args.type != "wds":
        raise cli.ValidationError("--no-cls requires --type wds to be set")

    if args.sampling_file is not None and len(args.data_path) > 0:
        raise cli.ValidationError("--sampling-file cannot be used with --data-path")
    if args.sampling_file is not None and args.target_path is None:
        raise cli.ValidationError("--sampling-file requires --target-path to be set")

    args.max_size = args.max_size * 1e6
    if args.target_path is None:
        pack_path = Path(f"{Path(args.data_path[0])}_{args.suffix}")
    else:
        pack_path = Path(args.target_path)

    if pack_path.exists() is False:
        logger.info(f"Creating {pack_path} directory...")
        pack_path.mkdir(parents=True)

    elif args.append is False:
        logger.warning("Directory already exists... aborting")
        raise SystemExit(1)

    pack(args, pack_path)



import random
from collections.abc import Callable
from typing import Any
from typing import Literal
from typing import Optional

import torch
from torch import nn
from torchvision.transforms import v2

from birder.data.transforms.classification import RGBType


class ResizeWithRandomInterpolation(nn.Module):
    def __init__(
        self, size: Optional[int] | tuple[int, int], max_size: Optional[int], interpolation: list[v2.InterpolationMode]
    ) -> None:
        super().__init__()
        self.transform = []
        for interp in interpolation:
            self.transform.append(
                v2.Resize(
                    size,
                    interpolation=interp,
                    max_size=max_size,
                    antialias=True,
                )
            )

    def forward(self, *x: Any) -> torch.Tensor:
        t = random.choice(self.transform)
        return t(x)


def get_birder_augment(
    size: tuple[int, int],
    level: int,
    fill_value: list[float],
    dynamic_size: bool,
    multiscale: bool,
    max_size: Optional[int],
) -> Callable[..., torch.Tensor]:
    if dynamic_size is True:
        target_size: Optional[int] | tuple[int, int] = min(size)
    elif max_size is not None:
        target_size = min(size)
    else:
        target_size = size

    transformations = []

    # Base augmentations
    if level >= 1:
        if dynamic_size is False and multiscale is False:
            transformations.extend(
                [
                    v2.RandomChoice(
                        [
                            v2.ScaleJitter(
                                target_size=size, scale_range=(max(0.1, 0.5 - (0.08 * level)), 2), antialias=True
                            ),
                            v2.RandomZoomOut(fill_value, side_range=(1, 3 + level * 0.1), p=0.5),
                        ]
                    ),
                ]
            )

    if level >= 3:
        transformations.extend(
            [
                v2.RandomIoUCrop(),
                v2.ClampBoundingBoxes(),
            ]
        )

    # Resize
    if multiscale is True:
        transformations.append(
            v2.RandomShortestSize(
                min_size=(480, 512, 544, 576, 608, 640, 672, 704, 736, 768, 800), max_size=max_size or 1333
            ),
        )
    else:
        transformations.append(
            ResizeWithRandomInterpolation(
                target_size, max_size, interpolation=[v2.InterpolationMode.BILINEAR, v2.InterpolationMode.BICUBIC]
            ),
        )

    # Classification style augmentations
    if level >= 4:
        transformations.extend(
            [
                v2.RandomChoice(
                    [
                        v2.ColorJitter(
                            brightness=0.1 + (0.0125 * level),
                            contrast=0.0 + (0.015 * level),
                            hue=max(0, -0.025 + (level * 0.01)),
                        ),
                        v2.RandomPhotometricDistort(p=1.0),
                        v2.Identity(),
                    ]
                ),
            ]
        )

    if level >= 6:
        transformations.extend(
            [
                v2.RandomChoice(
                    [
                        v2.RandomGrayscale(p=0.5),
                        v2.RandomSolarize(255 - (10 * level), p=0.5),
                    ]
                ),
            ]
        )

    transformations.extend(
        [
            v2.RandomHorizontalFlip(0.5),
            v2.SanitizeBoundingBoxes(),
        ]
    )

    return v2.Compose(transformations)  # type: ignore


AugType = Literal["birder", "lsj", "multiscale", "ssd", "ssdlite", "detr"]


def training_preset(
    size: tuple[int, int],
    aug_type: AugType,
    level: int,
    rgv_values: RGBType,
    dynamic_size: bool = False,
    multiscale: bool = False,
    max_size: Optional[int] = None,
) -> Callable[..., torch.Tensor]:
    mean = rgv_values["mean"]
    std = rgv_values["std"]
    fill_value = [255 * v for v in mean]
    if dynamic_size is True:
        target_size: Optional[int] | tuple[int, int] = min(size)
    elif max_size is not None:
        target_size = min(size)
    else:
        target_size = size

    if aug_type == "birder":
        if 0 > level or level > 10:
            raise ValueError("Unsupported aug level")

        return v2.Compose(  # type:ignore
            [
                v2.ToImage(),
                get_birder_augment(size, level, fill_value, dynamic_size, multiscale, max_size),
                v2.ToDtype(torch.float32, scale=True),
                v2.Normalize(mean=mean, std=std),
                v2.ToPureTensor(),
            ]
        )

    if aug_type == "lsj":
        return v2.Compose(  # type: ignore
            [
                v2.ToImage(),
                v2.ScaleJitter(target_size=size, scale_range=(0.1, 2), antialias=True),
                ResizeWithRandomInterpolation(  # Supposed to be FixedSizeCrop
                    target_size, max_size, interpolation=[v2.InterpolationMode.BILINEAR, v2.InterpolationMode.BICUBIC]
                ),
                v2.RandomHorizontalFlip(0.5),
                v2.SanitizeBoundingBoxes(),
                v2.ToDtype(torch.float32, scale=True),
                v2.Normalize(mean=mean, std=std),
                v2.ToPureTensor(),
            ]
        )

    if aug_type == "multiscale":
        return v2.Compose(  # type: ignore
            [
                v2.ToImage(),
                v2.RandomShortestSize(
                    min_size=(480, 512, 544, 576, 608, 640, 672, 704, 736, 768, 800), max_size=max_size or 1333
                ),
                v2.RandomHorizontalFlip(0.5),
                v2.SanitizeBoundingBoxes(),
                v2.ToDtype(torch.float32, scale=True),
                v2.Normalize(mean=mean, std=std),
                v2.ToPureTensor(),
            ]
        )

    if aug_type == "ssd":
        return v2.Compose(  # type: ignore
            [
                v2.ToImage(),
                v2.RandomPhotometricDistort(),
                v2.RandomZoomOut(fill_value),
                v2.RandomIoUCrop(),
                ResizeWithRandomInterpolation(
                    target_size, max_size, interpolation=[v2.InterpolationMode.BILINEAR, v2.InterpolationMode.BICUBIC]
                ),
                v2.RandomHorizontalFlip(0.5),
                v2.SanitizeBoundingBoxes(),
                v2.ToDtype(torch.float32, scale=True),
                v2.Normalize(mean=mean, std=std),
                v2.ToPureTensor(),
            ]
        )

    if aug_type == "ssdlite":
        return v2.Compose(  # type: ignore
            [
                v2.ToImage(),
                v2.RandomIoUCrop(),
                ResizeWithRandomInterpolation(
                    target_size, max_size, interpolation=[v2.InterpolationMode.BILINEAR, v2.InterpolationMode.BICUBIC]
                ),
                v2.RandomHorizontalFlip(0.5),
                v2.SanitizeBoundingBoxes(),
                v2.ToDtype(torch.float32, scale=True),
                v2.Normalize(mean=mean, std=std),
                v2.ToPureTensor(),
            ]
        )

    if aug_type == "detr":
        return v2.Compose(  # type: ignore
            [
                v2.ToImage(),
                v2.RandomChoice(
                    [
                        v2.RandomShortestSize(
                            (480, 512, 544, 576, 608, 640, 672, 704, 736, 768, 800), max_size=max_size or 1333
                        ),
                        v2.Compose(
                            [
                                v2.RandomShortestSize((400, 500, 600)),
                                v2.RandomIoUCrop(),  # Supposed to be RandomSizeCrop
                                v2.RandomShortestSize(
                                    (480, 512, 544, 576, 608, 640, 672, 704, 736, 768, 800), max_size=max_size or 1333
                                ),
                            ]
                        ),
                    ]
                ),
                v2.RandomHorizontalFlip(0.5),
                v2.SanitizeBoundingBoxes(),
                v2.ToDtype(torch.float32, scale=True),
                v2.Normalize(mean=mean, std=std),
                v2.ToPureTensor(),
            ]
        )

    raise ValueError("Unsupported augmentation type")


class InferenceTransform:
    def __init__(
        self,
        size: tuple[int, int],
        rgv_values: RGBType,
        dynamic_size: bool = False,
        max_size: Optional[int] = None,
        no_resize: bool = False,
    ):
        """
        Create a torchvision transform pipeline for detection inference

        This function builds a standardized preprocessing pipeline that converts input images
        to tensors with proper normalization and resizing for detection model inference.
        The pipeline handles various sizing strategies.

        Parameters
        ----------
        size
            Target image dimensions as (height, width). Behavior depends on other parameters:
            - With dynamic_size=False and max_size=None: Images resized exactly to this size
            - With dynamic_size=True: min(size) used as target for shorter edge, aspect ratio preserved
            - With max_size specified: Ignored in favor of max_size-based scaling
        rgv_values
            RGB normalization statistics containing 'mean' and 'std' tuples.
            Typically obtained from get_rgb_stats().
        dynamic_size
            When True, preserves aspect ratios by using min(size) as the target
            for the shorter edge. Longer edge scales proportionally.
            Respects max_size is specified.
        max_size
            Maximum allowed size for the longer edge.
        no_resize
            When True, skips resizing step entirely.
        """

        mean = rgv_values["mean"]
        std = rgv_values["std"]
        if dynamic_size is True:
            target_size: Optional[int] | tuple[int, int] = min(size)
        elif max_size is not None:
            target_size = min(size)
        else:
            target_size = size

        if no_resize is True:
            resize = v2.Identity()
        else:
            resize = v2.Resize(
                target_size, interpolation=v2.InterpolationMode.BICUBIC, max_size=max_size, antialias=True
            )

        self.transform = v2.Compose(
            [
                v2.ToImage(),
                resize,
                v2.ToDtype(torch.float32, scale=True),
                v2.Normalize(mean=mean, std=std),
                v2.ToPureTensor(),
            ]
        )

    def __call__(self, *inpt: Any, **kwargs: Any) -> Any:
        return self.transform(*inpt, **kwargs)

    @staticmethod
    def postprocess(
        detections: list[dict[str, torch.Tensor]], image_sizes: list[list[int]], original_image_sizes: list[list[int]]
    ) -> list[dict[str, torch.Tensor]]:
        for i, (detection, image_size, original_size) in enumerate(zip(detections, image_sizes, original_image_sizes)):
            if "boxes" not in detection:
                continue

            boxes = detection["boxes"]

            (orig_h, orig_w) = original_size
            h_ratio = orig_h / image_size[0]
            w_ratio = orig_w / image_size[1]
            adjusted_boxes = boxes * torch.tensor([w_ratio, h_ratio, w_ratio, h_ratio], device=boxes.device)

            detections[i]["boxes"] = adjusted_boxes

        return detections



import math
from collections import OrderedDict
from collections.abc import Callable
from typing import Any
from typing import Optional
from typing import TypedDict

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import FeaturePyramidNetwork
from torchvision.ops.feature_pyramid_network import ExtraFPNBlock

from birder.model_registry import Task
from birder.model_registry import registry
from birder.net.base import DataShapeType
from birder.net.base import DetectorBackbone

DetectorResultType = TypedDict(
    "DetectorResultType", {"boxes": torch.Tensor, "labels": torch.Tensor, "scores": torch.Tensor}
)
DetectionSignatureType = TypedDict(
    "DetectionSignatureType",
    {
        "dynamic": bool,
        "inputs": list[DataShapeType],
        "outputs": tuple[list[DetectorResultType], dict[str, torch.Tensor]],
        "num_labels": int,
    },
)


def get_detection_signature(input_shape: tuple[int, ...], num_outputs: int, dynamic: bool) -> DetectionSignatureType:
    return {
        "dynamic": dynamic,
        "inputs": [{"data_shape": [0, *input_shape[1:]]}],
        "outputs": ([{"boxes": [0, 4], "labels": [0], "scores": [0]}], {}),
        "num_labels": num_outputs,
    }


class DetectionBaseNet(nn.Module):
    default_size: tuple[int, int]
    auto_register = False
    scriptable = True
    task = str(Task.OBJECT_DETECTION)

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        if cls.auto_register is False:
            # Exclude networks with custom config (initialized only with aliases)
            return

        registry.register_model(cls.__name__.lower(), cls)

    def __init__(
        self,
        num_classes: int,
        backbone: DetectorBackbone,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
        export_mode: bool = False,
    ) -> None:
        super().__init__()
        self.input_channels = backbone.input_channels
        self.num_classes = num_classes + 1  # Background always at index 0
        self.backbone = backbone
        self.backbone.transform_to_backbone()
        if hasattr(self, "config") is False:  # Avoid overriding aliases
            self.config = config
        elif config is not None:
            assert self.config is not None
            self.config.update(config)  # Override with custom config

        if size is not None:
            self.size = size
        else:
            self.size = self.default_size

        assert isinstance(self.size, tuple)
        assert isinstance(self.size[0], int)
        assert isinstance(self.size[1], int)

        self.export_mode = export_mode
        self.dynamic_size = False

    def reset_classifier(self, num_classes: int) -> None:
        raise NotImplementedError

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        """
        Override this when one time adjustments for different resolutions is required.
        This should run after load_state_dict.
        """

        self.size = new_size
        self.backbone.adjust_size(new_size)

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:
        self.backbone.set_dynamic_size(dynamic_size)
        self.dynamic_size = dynamic_size

    def freeze(self, freeze_classifier: bool = True) -> None:
        raise NotImplementedError

    # pylint: disable=protected-access
    def _input_check(self, targets: Optional[list[dict[str, torch.Tensor]]]) -> None:
        if self.training is True:
            if targets is None:
                torch._assert(False, "targets should not be None when in training mode")

        if targets is not None:
            for target_idx, target in enumerate(targets):
                boxes = target["boxes"]
                if isinstance(boxes, torch.Tensor):
                    torch._assert(
                        len(boxes.shape) == 2 and boxes.shape[-1] == 4,
                        f"Expected target boxes to be a tensor of shape [N, 4], got {boxes.shape}.",
                    )

                else:
                    torch._assert(False, f"Expected target boxes to be of type Tensor, got {type(boxes)}.")

                degenerate_boxes = boxes[:, 2:] <= boxes[:, :2]
                if degenerate_boxes.any():
                    # Print the first degenerate box
                    bb_idx = torch.where(degenerate_boxes.any(dim=1))[0][0]
                    degenerate_bb: list[float] = boxes[bb_idx].tolist()
                    torch._assert(
                        False,
                        "All bounding boxes should have positive height and width."
                        f" Found invalid box {degenerate_bb} for target at index {target_idx}.",
                    )

    # pylint: disable=protected-access
    def _to_img_list(self, x: torch.Tensor, image_sizes: Optional[list[list[int]]] = None) -> "ImageList":
        if image_sizes is None:
            image_sizes = [img.shape[-2:] for img in x]

        image_sizes_list: list[tuple[int, int]] = []
        for image_size in image_sizes:
            torch._assert(
                len(image_size) == 2,
                f"Input tensors expected to have in the last two elements H and W, instead got {image_size}",
            )
            image_sizes_list.append((image_size[0], image_size[1]))

        return ImageList(x, image_sizes_list)

    def forward(
        self,
        x: torch.Tensor,
        targets: Optional[list[dict[str, torch.Tensor]]] = None,
        masks: Optional[torch.Tensor] = None,
        image_sizes: Optional[list[list[int]]] = None,
    ) -> tuple[list[dict[str, torch.Tensor]], dict[str, torch.Tensor]]:
        # TypedDict not supported for TorchScript - avoid returning DetectorResultType
        raise NotImplementedError


class ImageList:
    def __init__(self, tensors: torch.Tensor, image_sizes: list[tuple[int, int]]) -> None:
        self.tensors = tensors
        self.image_sizes = image_sizes

    def to(self, device: torch.device) -> "ImageList":
        cast_tensor = self.tensors.to(device)
        return ImageList(cast_tensor, self.image_sizes)


###############################################################################
# Backbone Classes
###############################################################################


class BackboneWithFPN(nn.Module):
    def __init__(
        self, backbone: DetectorBackbone, out_channels: int, extra_blocks: Optional[ExtraFPNBlock] = None
    ) -> None:
        super().__init__()
        self.backbone = backbone
        self.size = self.backbone.size
        self.fpn = FeaturePyramidNetwork(
            in_channels_list=self.backbone.return_channels,
            out_channels=out_channels,
            extra_blocks=extra_blocks,
            norm_layer=nn.BatchNorm2d,
        )

        self.out_channels = out_channels

    def forward(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        out: dict[str, torch.Tensor] = self.backbone.detection_features(x)
        out = self.fpn(out)
        return out


# Code adapted from https://github.com/fizyr-forks/torchvision/blob/vitdet2/torchvision/ops/feature_pyramid_network.py
# Reference license: BSD 3-Clause


class BackboneWithSimpleFPN(nn.Module):
    def __init__(
        self,
        backbone: DetectorBackbone,
        out_channels: int,
        extra_blocks: Optional[ExtraFPNBlock] = None,
        num_stages: int = 4,
    ) -> None:
        super().__init__()
        self.backbone = backbone
        self.size = self.backbone.size
        self.fpn = SimpleFeaturePyramidNetwork(
            in_channels=self.backbone.return_channels[-1],
            out_channels=out_channels,
            norm_layer=nn.BatchNorm2d,
            extra_blocks=extra_blocks,
            num_stages=num_stages,
        )

        self.out_channels = out_channels

    def forward(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        out: dict[str, torch.Tensor] = self.backbone.detection_features(x)
        x = out[self.backbone.return_stages[-1]]
        out = self.fpn(x)
        return out


class SimpleFeaturePyramidNetwork(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        norm_layer: Callable[..., nn.Module],
        extra_blocks: Optional[ExtraFPNBlock] = None,
        num_stages: int = 4,
    ):
        super().__init__()
        self.blocks = nn.ModuleList()
        for block_index in range(4 - num_stages, 4):
            layers = []

            current_in_channels = in_channels
            if block_index == 0:
                layers.extend(
                    [
                        nn.ConvTranspose2d(
                            in_channels, in_channels // 2, kernel_size=(2, 2), stride=(2, 2), padding=(0, 0)
                        ),
                        norm_layer(in_channels // 2),
                        nn.GELU(),
                        nn.ConvTranspose2d(
                            in_channels // 2, in_channels // 4, kernel_size=(2, 2), stride=(2, 2), padding=(0, 0)
                        ),
                    ]
                )
                current_in_channels = in_channels // 4
            elif block_index == 1:
                layers.append(
                    nn.ConvTranspose2d(
                        in_channels, in_channels // 2, kernel_size=(2, 2), stride=(2, 2), padding=(0, 0)
                    ),
                )
                current_in_channels = in_channels // 2
            elif block_index == 2:
                pass
            elif block_index == 3:
                layers.append(nn.MaxPool2d(kernel_size=(2, 2), stride=(2, 2)))

            layers.extend(
                [
                    Conv2dNormActivation(
                        current_in_channels,
                        out_channels,
                        kernel_size=(1, 1),
                        stride=(1, 1),
                        padding=(0, 0),
                        norm_layer=norm_layer,
                        activation_layer=None,
                    ),
                    Conv2dNormActivation(
                        out_channels,
                        out_channels,
                        kernel_size=(3, 3),
                        stride=(1, 1),
                        padding=(1, 1),
                        norm_layer=norm_layer,
                        activation_layer=None,
                    ),
                ]
            )

            self.blocks.append(nn.Sequential(*layers))

        if extra_blocks is not None:
            if not isinstance(extra_blocks, ExtraFPNBlock):
                raise TypeError(f"extra_blocks should be of type ExtraFPNBlock not {type(extra_blocks)}")

        self.extra_blocks = extra_blocks

    def forward(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        results: list[torch.Tensor] = []
        names: list[str] = []
        for idx, block in enumerate(self.blocks):
            results.append(block(x))
            names.append(f"stage{idx+1}")

        if self.extra_blocks is not None:
            (results, names) = self.extra_blocks(results, [x], names)

        out = OrderedDict(list(zip(names, results)))

        return out


###############################################################################
# General Detection
###############################################################################

# Code adapted from https://github.com/pytorch/vision/blob/main/torchvision/models/detection/_utils.py
# Reference license: BSD 3-Clause


# pylint: disable=protected-access,too-many-locals
@torch.jit._script_if_tracing  # type: ignore
def encode_boxes(reference_boxes: torch.Tensor, proposals: torch.Tensor, weights: torch.Tensor) -> torch.Tensor:
    """
    Encode a set of proposals with respect to some reference boxes
    """

    # Perform some unpacking to make it JIT-fusion friendly
    wx = weights[0]
    wy = weights[1]
    ww = weights[2]
    wh = weights[3]

    proposals_x1 = proposals[:, 0].unsqueeze(1)
    proposals_y1 = proposals[:, 1].unsqueeze(1)
    proposals_x2 = proposals[:, 2].unsqueeze(1)
    proposals_y2 = proposals[:, 3].unsqueeze(1)

    reference_boxes_x1 = reference_boxes[:, 0].unsqueeze(1)
    reference_boxes_y1 = reference_boxes[:, 1].unsqueeze(1)
    reference_boxes_x2 = reference_boxes[:, 2].unsqueeze(1)
    reference_boxes_y2 = reference_boxes[:, 3].unsqueeze(1)

    # Implementation starts here
    ex_widths = proposals_x2 - proposals_x1
    ex_heights = proposals_y2 - proposals_y1
    ex_ctr_x = proposals_x1 + 0.5 * ex_widths
    ex_ctr_y = proposals_y1 + 0.5 * ex_heights

    gt_widths = reference_boxes_x2 - reference_boxes_x1
    gt_heights = reference_boxes_y2 - reference_boxes_y1
    gt_ctr_x = reference_boxes_x1 + 0.5 * gt_widths
    gt_ctr_y = reference_boxes_y1 + 0.5 * gt_heights

    targets_dx = wx * (gt_ctr_x - ex_ctr_x) / ex_widths
    targets_dy = wy * (gt_ctr_y - ex_ctr_y) / ex_heights
    targets_dw = ww * torch.log(gt_widths / ex_widths)
    targets_dh = wh * torch.log(gt_heights / ex_heights)

    targets = torch.concat((targets_dx, targets_dy, targets_dw, targets_dh), dim=1)

    return targets


class BoxCoder:
    """
    This class encodes and decodes a set of bounding boxes into
    the representation used for training the regressors.
    """

    def __init__(
        self, weights: tuple[float, float, float, float], bbox_xform_clip: float = math.log(1000.0 / 16)
    ) -> None:
        self.weights = weights
        self.bbox_xform_clip = bbox_xform_clip

    def encode(self, reference_boxes: list[torch.Tensor], proposals: list[torch.Tensor]) -> list[torch.Tensor]:
        boxes_per_image = [len(b) for b in reference_boxes]
        reference_boxes = torch.concat(reference_boxes, dim=0)
        proposals = torch.concat(proposals, dim=0)
        targets = self.encode_single(reference_boxes, proposals)
        target_list: list[torch.Tensor] = targets.split(boxes_per_image, 0)

        return target_list

    def encode_single(self, reference_boxes: torch.Tensor, proposals: torch.Tensor) -> torch.Tensor:
        """
        Encode a set of proposals with respect to some reference boxes
        """

        dtype = reference_boxes.dtype
        device = reference_boxes.device
        weights = torch.as_tensor(self.weights, dtype=dtype, device=device)
        targets = encode_boxes(reference_boxes, proposals, weights)

        return targets

    # pylint: disable=protected-access
    def decode(self, rel_codes: torch.Tensor, boxes: list[torch.Tensor]) -> torch.Tensor:
        torch._assert(isinstance(boxes, (list, tuple)), "This function expects boxes of type list or tuple.")
        torch._assert(isinstance(rel_codes, torch.Tensor), "This function expects rel_codes of type torch.Tensor.")

        boxes_per_image = [b.size(0) for b in boxes]
        concat_boxes = torch.concat(boxes, dim=0)
        box_sum = 0
        for val in boxes_per_image:
            box_sum += val

        if box_sum > 0:
            rel_codes = rel_codes.reshape(box_sum, -1)

        pred_boxes = self.decode_single(rel_codes, concat_boxes)
        if box_sum > 0:
            pred_boxes = pred_boxes.reshape(box_sum, -1, 4)

        return pred_boxes

    def decode_single(self, rel_codes: torch.Tensor, boxes: torch.Tensor) -> torch.Tensor:
        """
        From a set of original boxes and encoded relative box offsets, get the decoded boxes
        """

        boxes = boxes.to(rel_codes.dtype)

        widths = boxes[:, 2] - boxes[:, 0]
        heights = boxes[:, 3] - boxes[:, 1]
        ctr_x = boxes[:, 0] + 0.5 * widths
        ctr_y = boxes[:, 1] + 0.5 * heights

        (wx, wy, ww, wh) = self.weights
        dx = rel_codes[:, 0::4] / wx
        dy = rel_codes[:, 1::4] / wy
        dw = rel_codes[:, 2::4] / ww
        dh = rel_codes[:, 3::4] / wh

        # Prevent sending too large values into torch.exp()
        dw = torch.clamp(dw, max=self.bbox_xform_clip)
        dh = torch.clamp(dh, max=self.bbox_xform_clip)

        pred_ctr_x = dx * widths[:, None] + ctr_x[:, None]
        pred_ctr_y = dy * heights[:, None] + ctr_y[:, None]
        pred_w = torch.exp(dw) * widths[:, None]
        pred_h = torch.exp(dh) * heights[:, None]

        # Distance from center to box's corner.
        c_to_c_h = torch.tensor(0.5, dtype=pred_ctr_y.dtype, device=pred_h.device) * pred_h
        c_to_c_w = torch.tensor(0.5, dtype=pred_ctr_x.dtype, device=pred_w.device) * pred_w

        pred_boxes1 = pred_ctr_x - c_to_c_w
        pred_boxes2 = pred_ctr_y - c_to_c_h
        pred_boxes3 = pred_ctr_x + c_to_c_w
        pred_boxes4 = pred_ctr_y + c_to_c_h
        pred_boxes = torch.stack((pred_boxes1, pred_boxes2, pred_boxes3, pred_boxes4), dim=2).flatten(1)

        return pred_boxes


class AnchorGenerator(nn.Module):
    def __init__(
        self,
        sizes: list[list[int]],
        aspect_ratios: list[list[float]],
    ) -> None:
        super().__init__()
        self.sizes = sizes
        self.aspect_ratios = aspect_ratios
        self.cell_anchors = [
            self.generate_anchors(size, aspect_ratio) for size, aspect_ratio in zip(sizes, aspect_ratios)
        ]

    def generate_anchors(
        self,
        scales_list: list[int],
        aspect_ratios_list: list[float],
        dtype: torch.dtype = torch.float32,
    ) -> torch.Tensor:
        scales = torch.as_tensor(scales_list, dtype=dtype)
        aspect_ratios = torch.as_tensor(aspect_ratios_list, dtype=dtype)
        h_ratios = torch.sqrt(aspect_ratios)
        w_ratios = 1 / h_ratios

        ws = (w_ratios[:, None] * scales[None, :]).view(-1)
        hs = (h_ratios[:, None] * scales[None, :]).view(-1)

        base_anchors = torch.stack([-ws, -hs, ws, hs], dim=1) / 2

        return base_anchors.round()

    def num_anchors_per_location(self) -> list[int]:
        return [len(s) * len(a) for s, a in zip(self.sizes, self.aspect_ratios)]

    # For every combination of (a, (g, s), i) in (self.cell_anchors, zip(grid_sizes, strides), 0:2),
    # output g[i] anchors that are s[i] distance apart in direction i, with the same dimensions as a.
    # pylint: disable=protected-access
    def grid_anchors(self, grid_sizes: list[list[int]], strides: list[list[torch.Tensor]]) -> list[torch.Tensor]:
        anchors = []
        cell_anchors = self.cell_anchors
        torch._assert(cell_anchors is not None, "cell_anchors should not be None")
        torch._assert(
            len(grid_sizes) == len(strides) == len(cell_anchors),
            "Anchors should be Tuple[Tuple[int]] because each feature "
            "map could potentially have different sizes and aspect ratios. "
            "There needs to be a match between the number of "
            "feature maps passed and the number of sizes / aspect ratios specified.",
        )

        for size, stride, base_anchors in zip(grid_sizes, strides, cell_anchors):
            (grid_height, grid_width) = size
            (stride_height, stride_width) = stride
            device = base_anchors.device

            # For output anchor, compute [x_center, y_center, x_center, y_center]
            shifts_x = torch.arange(0, grid_width, dtype=torch.int32, device=device) * stride_width
            shifts_y = torch.arange(0, grid_height, dtype=torch.int32, device=device) * stride_height
            shift_y, shift_x = torch.meshgrid(shifts_y, shifts_x, indexing="ij")
            shift_x = shift_x.reshape(-1)
            shift_y = shift_y.reshape(-1)
            shifts = torch.stack((shift_x, shift_y, shift_x, shift_y), dim=1)

            # For every (base anchor, output anchor) pair,
            # offset each zero-centered base anchor by the center of the output anchor.
            anchors.append((shifts.view(-1, 1, 4) + base_anchors.view(1, -1, 4)).reshape(-1, 4))

        return anchors

    def forward(self, image_list: ImageList, feature_maps: list[torch.Tensor]) -> list[torch.Tensor]:
        grid_sizes = [feature_map.shape[-2:] for feature_map in feature_maps]
        image_size = image_list.tensors.shape[-2:]
        dtype = feature_maps[0].dtype
        device = feature_maps[0].device
        strides = [
            [
                torch.empty((), dtype=torch.int64, device=device).fill_(image_size[0] // g[0]),
                torch.empty((), dtype=torch.int64, device=device).fill_(image_size[1] // g[1]),
            ]
            for g in grid_sizes
        ]
        self.cell_anchors = [cell_anchor.to(dtype=dtype, device=device) for cell_anchor in self.cell_anchors]
        anchors_over_all_feature_maps = self.grid_anchors(grid_sizes, strides)
        anchors: list[list[torch.Tensor]] = []
        for _ in range(len(image_list.image_sizes)):
            anchors.append(anchors_over_all_feature_maps)

        anchors = [torch.concat(anchors_per_image, dim=0) for anchors_per_image in anchors]
        return anchors


class Matcher(nn.Module):
    """
    This class assigns to each predicted "element" (e.g., a box) a ground-truth
    element. Each predicted element will have exactly zero or one matches; each
    ground-truth element may be assigned to zero or more predicted elements.

    Matching is based on the MxN match_quality_matrix, that characterizes how well
    each (ground-truth, predicted)-pair match. For example, if the elements are
    boxes, the matrix may contain box IoU overlap values.

    The matcher returns a tensor of size N containing the index of the ground-truth
    element m that matches to prediction n. If there is no match, a negative value
    is returned.
    """

    BELOW_LOW_THRESHOLD = -1
    BETWEEN_THRESHOLDS = -2

    def __init__(self, high_threshold: float, low_threshold: float, allow_low_quality_matches: bool = False) -> None:
        """
        Parameters
        ----------
        high_threshold
            quality values greater than or equal to this value are candidate matches
        low_threshold
            a lower quality threshold used to stratify matches into three levels:
            1) matches >= high_threshold
            2) BETWEEN_THRESHOLDS matches in [low_threshold, high_threshold)
            3) BELOW_LOW_THRESHOLD matches in [0, low_threshold)
        allow_low_quality_matches
            if True, produce additional matches
            for predictions that have only low-quality match candidates.
        """

        super().__init__()
        self.BELOW_LOW_THRESHOLD = Matcher.BELOW_LOW_THRESHOLD  # pylint: disable=invalid-name
        self.BETWEEN_THRESHOLDS = Matcher.BETWEEN_THRESHOLDS  # pylint: disable=invalid-name
        torch._assert(low_threshold <= high_threshold, "low_threshold should be <= high_threshold")
        self.high_threshold = high_threshold
        self.low_threshold = low_threshold
        self.allow_low_quality_matches = allow_low_quality_matches

    def forward(self, match_quality_matrix: torch.Tensor) -> torch.Tensor:
        """
        Parameters
        ----------
        match_quality_matrix
            an MxN tensor, containing the pairwise quality
            between M ground-truth elements and N predicted elements

        Returns
        -------
        an N tensor where N[i] is a matched gt in [0, M - 1] or a negative
        value indicating that prediction i could not be matched
        """

        if match_quality_matrix.numel() == 0:
            # Empty targets or proposals not supported during training
            if match_quality_matrix.shape[0] == 0:
                raise ValueError("No ground-truth boxes available for one of the images during training")

            raise ValueError("No proposal boxes available for one of the images during training")

        # match_quality_matrix is M (gt) x N (predicted)
        # Max over gt elements (dim 0) to find best gt candidate for each prediction
        matched_vals, matches = match_quality_matrix.max(dim=0)
        if self.allow_low_quality_matches is True:
            all_matches = matches.clone()
        else:
            all_matches = None

        # Assign candidate matches with low quality to negative (unassigned) values
        below_low_threshold = matched_vals < self.low_threshold
        between_thresholds = (matched_vals >= self.low_threshold) & (matched_vals < self.high_threshold)
        matches[below_low_threshold] = self.BELOW_LOW_THRESHOLD
        matches[between_thresholds] = self.BETWEEN_THRESHOLDS

        if self.allow_low_quality_matches is True:
            if all_matches is None:
                torch._assert(False, "all_matches should not be None")  # pylint: disable=protected-access
            else:
                self.set_low_quality_matches_(matches, all_matches, match_quality_matrix)

        return matches

    def set_low_quality_matches_(
        self, matches: torch.Tensor, all_matches: torch.Tensor, match_quality_matrix: torch.Tensor
    ) -> None:
        """
        Produce additional matches for predictions that have only low-quality matches.
        Specifically, for each ground-truth find the set of predictions that have
        maximum overlap with it (including ties); for each prediction in that set, if
        it is unmatched, then match it to the ground-truth with which it has the highest
        quality value.
        """

        # For each gt, find the prediction with which it has the highest quality
        highest_quality_foreach_gt, _ = match_quality_matrix.max(dim=1)

        # Find the highest quality match available, even if it is low, including ties
        gt_pred_pairs_of_highest_quality = torch.where(match_quality_matrix == highest_quality_foreach_gt[:, None])
        # Example gt_pred_pairs_of_highest_quality:
        # (tensor([0, 1, 1, 2, 2, 3, 3, 4, 5, 5]),
        #  tensor([39796, 32055, 32070, 39190, 40255, 40390, 41455, 45470, 45325, 46390]))
        # Each element in the first tensor is a gt index,
        # and each element in second tensor is a prediction index
        # Note how gt items 1, 2, 3, and 5 each have two ties

        pred_idx_to_update = gt_pred_pairs_of_highest_quality[1]
        matches[pred_idx_to_update] = all_matches[pred_idx_to_update]



"""
MobileNet v4, adapted from
https://github.com/jaiwei98/MobileNetV4-pytorch/blob/main/mobilenet/mobilenetv4.py

Paper "MobileNetV4 -- Universal Models for the Mobile Ecosystem", https://arxiv.org/abs/2404.10518
"""

# Reference license: MIT

from collections import OrderedDict
from collections.abc import Callable
from typing import Any
from typing import Literal
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import SqueezeExcitation
from torchvision.ops import StochasticDepth

from birder.common.masking import mask_tensor
from birder.layers import LayerScale2d
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenRetentionResultType
from birder.net.base import make_divisible


class ConvNormActConfig:
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel: tuple[int, int],
        stride: tuple[int, int],
        padding: tuple[int, int],
        activation: Callable[..., nn.Module] = nn.ReLU,
    ) -> None:
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel = kernel
        self.stride = stride
        self.padding = padding
        self.activation = activation


class InvertedResidualConfig:
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: tuple[int, int],
        stride: tuple[int, int],
        padding: tuple[int, int],
        width_multi: float,
        use_se: bool,
        activation: Callable[..., nn.Module] = nn.ReLU,
    ):
        self.in_channels = make_divisible(in_channels, 8)
        self.out_channels = make_divisible(out_channels, 8)
        self.expanded_channels = make_divisible(in_channels * width_multi, 8)
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.use_se = use_se
        self.activation = activation


class UniversalInvertedBottleneckConfig:
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        expand_ratio: float,
        start_dw_kernel_size: Optional[tuple[int, int]],
        middle_dw_kernel_size: Optional[tuple[int, int]],
        stride: tuple[int, int],
        middle_dw_downsample: bool,
        activation: Callable[..., nn.Module] = nn.ReLU,
    ) -> None:
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.expand_ratio = expand_ratio
        self.start_dw_kernel_size = start_dw_kernel_size
        self.middle_dw_kernel_size = middle_dw_kernel_size
        self.stride = stride
        self.middle_dw_downsample = middle_dw_downsample
        self.activation = activation

        if stride[0] == 1 and stride[1] == 1 and in_channels == out_channels:
            self.shortcut = True
        else:
            self.shortcut = False


class InvertedResidual(nn.Module):
    def __init__(self, cnf: InvertedResidualConfig, stochastic_depth_prob: float) -> None:
        super().__init__()
        if cnf.stride[0] == 1 and cnf.stride[1] == 1 and cnf.in_channels == cnf.out_channels:
            self.shortcut = True
        else:
            self.shortcut = False

        layers = []
        # Expand
        if cnf.expanded_channels != cnf.in_channels:
            layers.append(
                Conv2dNormActivation(
                    cnf.in_channels,
                    cnf.expanded_channels,
                    kernel_size=cnf.kernel_size,
                    stride=cnf.stride,
                    padding=cnf.padding,
                    activation_layer=cnf.activation,
                    inplace=None,
                    bias=False,
                )
            )

        if cnf.use_se is True:
            # Depthwise
            layers.append(
                Conv2dNormActivation(
                    cnf.expanded_channels,
                    cnf.expanded_channels,
                    kernel_size=cnf.kernel_size,
                    stride=cnf.stride,
                    padding=cnf.padding,
                    groups=cnf.expanded_channels,
                    activation_layer=cnf.activation,
                    inplace=None,
                    bias=False,
                )
            )
            squeeze_channels = make_divisible(cnf.expanded_channels // 4, 8)
            layers.append(SqueezeExcitation(cnf.expanded_channels, squeeze_channels, scale_activation=nn.Hardsigmoid))

        # Project
        layers.append(
            Conv2dNormActivation(
                cnf.expanded_channels,
                cnf.out_channels,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                bias=False,
                activation_layer=None,
            )
        )

        self.block = nn.Sequential(*layers)
        self.stochastic_depth = StochasticDepth(stochastic_depth_prob, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.shortcut is True:
            return x + self.stochastic_depth(self.block(x))

        return self.block(x)


class UniversalInvertedBottleneck(nn.Module):
    def __init__(
        self, cnf: UniversalInvertedBottleneckConfig, layer_scale_init_value: float, stochastic_depth_prob: float
    ) -> None:
        super().__init__()
        self.shortcut = cnf.shortcut

        if cnf.start_dw_kernel_size is not None:
            if cnf.middle_dw_downsample is True:
                s = (1, 1)
            else:
                s = cnf.stride

            self.start_dw_conv = Conv2dNormActivation(
                cnf.in_channels,
                cnf.in_channels,
                kernel_size=cnf.start_dw_kernel_size,
                stride=s,
                padding=((cnf.start_dw_kernel_size[0] - 1) // 2, (cnf.start_dw_kernel_size[1] - 1) // 2),
                groups=cnf.in_channels,
                activation_layer=None,
                bias=False,
            )

        else:
            self.start_dw_conv = nn.Identity()

        expand_channels = make_divisible(cnf.in_channels * cnf.expand_ratio, 8)
        self.expand_conv = Conv2dNormActivation(
            cnf.in_channels,
            expand_channels,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            activation_layer=cnf.activation,
            inplace=None,
            bias=False,
        )

        if cnf.middle_dw_kernel_size is not None:
            if cnf.middle_dw_downsample is True:
                s = cnf.stride
            else:
                s = (1, 1)

            self.middle_dw_conv = Conv2dNormActivation(
                expand_channels,
                expand_channels,
                kernel_size=cnf.middle_dw_kernel_size,
                stride=s,
                padding=((cnf.middle_dw_kernel_size[0] - 1) // 2, (cnf.middle_dw_kernel_size[1] - 1) // 2),
                groups=expand_channels,
                activation_layer=cnf.activation,
                inplace=None,
                bias=False,
            )

        else:
            self.middle_dw_conv = nn.Identity()

        self.proj_conv = Conv2dNormActivation(
            expand_channels,
            cnf.out_channels,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            activation_layer=None,
            bias=False,
        )

        if layer_scale_init_value > 0:
            self.layer_scale = LayerScale2d(cnf.out_channels, layer_scale_init_value)
        else:
            self.layer_scale = nn.Identity()

        self.stochastic_depth = StochasticDepth(stochastic_depth_prob, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        shortcut = x
        x = self.start_dw_conv(x)
        x = self.expand_conv(x)
        x = self.middle_dw_conv(x)
        x = self.proj_conv(x)
        x = self.layer_scale(x)
        if self.shortcut is True:
            return self.stochastic_depth(x) + shortcut

        return x


# pylint: disable=invalid-name
class MobileNet_v4(DetectorBackbone, PreTrainEncoder, MaskedTokenRetentionMixin):
    block_group_regex = r"body\.stage(\d+)\.(\d+)"

    # pylint: disable=too-many-branches
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        dropout: float = self.config["dropout"]
        drop_path_rate: float = self.config["drop_path_rate"]
        net_size: str = self.config["net_size"]

        if net_size == "small":
            stem_settings = ConvNormActConfig(0, 32, (3, 3), (2, 2), (1, 1))
            net_settings: list[Any] = [
                # Stage 1
                ConvNormActConfig(32, 32, (3, 3), (2, 2), (1, 1)),
                ConvNormActConfig(32, 32, (1, 1), (1, 1), (0, 0)),
                # Stage 2
                ConvNormActConfig(32, 96, (3, 3), (2, 2), (1, 1)),
                ConvNormActConfig(96, 64, (1, 1), (1, 1), (0, 0)),
                # Stage 3
                UniversalInvertedBottleneckConfig(64, 96, 3.0, (5, 5), (5, 5), (2, 2), True),
                UniversalInvertedBottleneckConfig(96, 96, 2.0, None, (3, 3), (1, 1), True),
                UniversalInvertedBottleneckConfig(96, 96, 2.0, None, (3, 3), (1, 1), True),
                UniversalInvertedBottleneckConfig(96, 96, 2.0, None, (3, 3), (1, 1), True),
                UniversalInvertedBottleneckConfig(96, 96, 2.0, None, (3, 3), (1, 1), True),
                UniversalInvertedBottleneckConfig(96, 96, 4.0, (3, 3), None, (1, 1), True),
                # Stage 4
                UniversalInvertedBottleneckConfig(96, 128, 6.0, (3, 3), (3, 3), (2, 2), True),
                UniversalInvertedBottleneckConfig(128, 128, 4.0, (5, 5), (5, 5), (1, 1), True),
                UniversalInvertedBottleneckConfig(128, 128, 4.0, None, (5, 5), (1, 1), True),
                UniversalInvertedBottleneckConfig(128, 128, 3.0, None, (5, 5), (1, 1), True),
                UniversalInvertedBottleneckConfig(128, 128, 4.0, None, (3, 3), (1, 1), True),
                UniversalInvertedBottleneckConfig(128, 128, 4.0, None, (3, 3), (1, 1), True),
            ]
            last_stage_settings = [
                ConvNormActConfig(128, 960, (1, 1), (1, 1), (0, 0)),
            ]
            features_stage_settings = ConvNormActConfig(960, 1280, (1, 1), (1, 1), (0, 0))
        elif net_size == "medium":
            stem_settings = ConvNormActConfig(0, 32, (3, 3), (2, 2), (1, 1))
            net_settings = [
                # Stage 1
                InvertedResidualConfig(32, 48, (3, 3), (2, 2), (1, 1), 4.0, False),
                # Stage 2
                UniversalInvertedBottleneckConfig(48, 80, 4.0, (3, 3), (5, 5), (2, 2), True),
                UniversalInvertedBottleneckConfig(80, 80, 2.0, (3, 3), (3, 3), (1, 1), True),
                # Stage 3
                UniversalInvertedBottleneckConfig(80, 160, 6.0, (3, 3), (5, 5), (2, 2), True),
                UniversalInvertedBottleneckConfig(160, 160, 4.0, (3, 3), (3, 3), (1, 1), True),
                UniversalInvertedBottleneckConfig(160, 160, 4.0, (3, 3), (3, 3), (1, 1), True),
                UniversalInvertedBottleneckConfig(160, 160, 4.0, (3, 3), (5, 5), (1, 1), True),
                UniversalInvertedBottleneckConfig(160, 160, 4.0, (3, 3), (3, 3), (1, 1), True),
                UniversalInvertedBottleneckConfig(160, 160, 4.0, (3, 3), None, (1, 1), True),
                UniversalInvertedBottleneckConfig(160, 160, 2.0, None, None, (1, 1), True),
                UniversalInvertedBottleneckConfig(160, 160, 4.0, (3, 3), None, (1, 1), True),
                # Stage 4
                UniversalInvertedBottleneckConfig(160, 256, 6.0, (5, 5), (5, 5), (2, 2), True),
                UniversalInvertedBottleneckConfig(256, 256, 4.0, (5, 5), (5, 5), (1, 1), True),
                UniversalInvertedBottleneckConfig(256, 256, 4.0, (3, 3), (5, 5), (1, 1), True),
                UniversalInvertedBottleneckConfig(256, 256, 4.0, (3, 3), (5, 5), (1, 1), True),
                UniversalInvertedBottleneckConfig(256, 256, 4.0, None, None, (1, 1), True),
                UniversalInvertedBottleneckConfig(256, 256, 4.0, (3, 3), None, (1, 1), True),
                UniversalInvertedBottleneckConfig(256, 256, 2.0, (3, 3), (5, 5), (1, 1), True),
                UniversalInvertedBottleneckConfig(256, 256, 4.0, (5, 5), (5, 5), (1, 1), True),
                UniversalInvertedBottleneckConfig(256, 256, 4.0, None, None, (1, 1), True),
                UniversalInvertedBottleneckConfig(256, 256, 4.0, None, None, (1, 1), True),
                UniversalInvertedBottleneckConfig(256, 256, 2.0, (5, 5), None, (1, 1), True),
            ]
            last_stage_settings = [
                ConvNormActConfig(256, 960, (1, 1), (1, 1), (0, 0)),
            ]
            features_stage_settings = ConvNormActConfig(960, 1280, (1, 1), (1, 1), (0, 0))
        elif net_size == "large":
            stem_settings = ConvNormActConfig(0, 24, (3, 3), (2, 2), (1, 1))
            net_settings = [
                # Stage 1
                InvertedResidualConfig(24, 48, (3, 3), (2, 2), (1, 1), 4.0, False),
                # Stage 2
                UniversalInvertedBottleneckConfig(48, 96, 4.0, (3, 3), (5, 5), (2, 2), True),
                UniversalInvertedBottleneckConfig(96, 96, 4.0, (3, 3), (3, 3), (1, 1), True),
                # Stage 3
                UniversalInvertedBottleneckConfig(96, 192, 4.0, (3, 3), (5, 5), (2, 2), True),
                UniversalInvertedBottleneckConfig(192, 192, 4.0, (3, 3), (3, 3), (1, 1), True),
                UniversalInvertedBottleneckConfig(192, 192, 4.0, (3, 3), (3, 3), (1, 1), True),
                UniversalInvertedBottleneckConfig(192, 192, 4.0, (3, 3), (3, 3), (1, 1), True),
                UniversalInvertedBottleneckConfig(192, 192, 4.0, (3, 3), (5, 5), (1, 1), True),
                UniversalInvertedBottleneckConfig(192, 192, 4.0, (5, 5), (3, 3), (1, 1), True),
                UniversalInvertedBottleneckConfig(192, 192, 4.0, (5, 5), (3, 3), (1, 1), True),
                UniversalInvertedBottleneckConfig(192, 192, 4.0, (5, 5), (3, 3), (1, 1), True),
                UniversalInvertedBottleneckConfig(192, 192, 4.0, (5, 5), (3, 3), (1, 1), True),
                UniversalInvertedBottleneckConfig(192, 192, 4.0, (5, 5), (3, 3), (1, 1), True),
                UniversalInvertedBottleneckConfig(192, 192, 4.0, (3, 3), None, (1, 1), True),
                # Stage 4
                UniversalInvertedBottleneckConfig(192, 512, 4.0, (5, 5), (5, 5), (2, 2), True),
                UniversalInvertedBottleneckConfig(512, 512, 4.0, (5, 5), (5, 5), (1, 1), True),
                UniversalInvertedBottleneckConfig(512, 512, 4.0, (5, 5), (5, 5), (1, 1), True),
                UniversalInvertedBottleneckConfig(512, 512, 4.0, (5, 5), (5, 5), (1, 1), True),
                UniversalInvertedBottleneckConfig(512, 512, 4.0, (5, 5), None, (1, 1), True),
                UniversalInvertedBottleneckConfig(512, 512, 4.0, (5, 5), (3, 3), (1, 1), True),
                UniversalInvertedBottleneckConfig(512, 512, 4.0, (5, 5), None, (1, 1), True),
                UniversalInvertedBottleneckConfig(512, 512, 4.0, (5, 5), None, (1, 1), True),
                UniversalInvertedBottleneckConfig(512, 512, 4.0, (5, 5), (3, 3), (1, 1), True),
                UniversalInvertedBottleneckConfig(512, 512, 4.0, (5, 5), (5, 5), (1, 1), True),
                UniversalInvertedBottleneckConfig(512, 512, 4.0, (5, 5), None, (1, 1), True),
                UniversalInvertedBottleneckConfig(512, 512, 4.0, (5, 5), None, (1, 1), True),
                UniversalInvertedBottleneckConfig(512, 512, 4.0, (5, 5), None, (1, 1), True),
            ]
            last_stage_settings = [
                ConvNormActConfig(512, 960, (1, 1), (1, 1), (0, 0)),
            ]
            features_stage_settings = ConvNormActConfig(960, 1280, (1, 1), (1, 1), (0, 0))
        else:
            raise ValueError(f"Unknown net_size '{net_size}'")

        self.stem = Conv2dNormActivation(
            self.input_channels,
            stem_settings.out_channels,
            kernel_size=stem_settings.kernel,
            stride=stem_settings.stride,
            padding=stem_settings.padding,
            activation_layer=stem_settings.activation,
        )

        layers: list[nn.Module] = []
        total_stage_blocks = len(net_settings) + len(last_stage_settings)
        i = 0
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for idx, block_settings in enumerate(net_settings):
            # Adjust stochastic depth probability based on the depth of the stage block
            sd_prob = drop_path_rate * float(idx) / total_stage_blocks

            if idx > 0 and (block_settings.stride[0] > 1 or block_settings.stride[1] > 1):
                stages[f"stage{i+1}"] = nn.Sequential(*layers)
                return_channels.append(net_settings[idx - 1].out_channels)
                layers = []
                i += 1

            if isinstance(block_settings, ConvNormActConfig):
                layers.append(
                    Conv2dNormActivation(
                        block_settings.in_channels,
                        block_settings.out_channels,
                        kernel_size=block_settings.kernel,
                        stride=block_settings.stride,
                        padding=block_settings.padding,
                        activation_layer=block_settings.activation,
                    )
                )

            elif isinstance(block_settings, InvertedResidualConfig):
                layers.append(InvertedResidual(block_settings, sd_prob))

            elif isinstance(block_settings, UniversalInvertedBottleneckConfig):
                layers.append(UniversalInvertedBottleneck(block_settings, 0.0, sd_prob))

            else:
                raise ValueError("Unknown config")

        stages[f"stage{i+1}"] = nn.Sequential(*layers)
        return_channels.append(net_settings[-1].out_channels)
        layers = []
        i += 1
        for block_settings in last_stage_settings:
            layers.append(
                Conv2dNormActivation(
                    block_settings.in_channels,
                    block_settings.out_channels,
                    kernel_size=block_settings.kernel,
                    stride=block_settings.stride,
                    padding=block_settings.padding,
                    activation_layer=block_settings.activation,
                )
            )
        stages[f"stage{i+1}"] = nn.Sequential(*layers)
        return_channels.append(last_stage_settings[-1].out_channels)

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            Conv2dNormActivation(
                features_stage_settings.in_channels,
                features_stage_settings.out_channels,
                kernel_size=features_stage_settings.kernel,
                stride=features_stage_settings.stride,
                padding=features_stage_settings.padding,
                activation_layer=features_stage_settings.activation,
            ),
            nn.Flatten(1),
            nn.Dropout(p=dropout),
        )
        self.return_channels = return_channels[:4]
        self.embedding_size = features_stage_settings.out_channels
        self.classifier = self.create_classifier()

        self.stem_stride = stem_settings.stride[0]
        self.stem_width = stem_settings.out_channels
        self.encoding_size = features_stage_settings.out_channels

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out")
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)

            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.zeros_(m.bias)

    def transform_to_backbone(self) -> None:
        self.body.stage5 = nn.Identity()
        self.features = nn.Identity()
        self.classifier = nn.Identity()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        x = self.stem(x)
        x = mask_tensor(x, mask, patch_factor=self.max_stride // self.stem_stride, mask_token=mask_token)
        x = self.body(x)

        result: TokenRetentionResultType = {}
        if return_keys in ("all", "features"):
            result["features"] = x
        if return_keys in ("all", "embedding"):
            result["embedding"] = self.features(x)

        return result

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config(
    "mobilenet_v4_s", MobileNet_v4, config={"dropout": 0.3, "drop_path_rate": 0.0, "net_size": "small"}
)
registry.register_model_config(
    "mobilenet_v4_m", MobileNet_v4, config={"dropout": 0.2, "drop_path_rate": 0.075, "net_size": "medium"}
)
registry.register_model_config(
    "mobilenet_v4_l", MobileNet_v4, config={"dropout": 0.2, "drop_path_rate": 0.35, "net_size": "large"}
)

registry.register_weights(
    "mobilenet_v4_s_il-common",
    {
        "description": "MobileNet v4 small model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 11.5,
                "sha256": "1b9b83a54c5be13e1fd6baec1f70524911cc3b7977b14c68ae45a85dc2fde911",
            }
        },
        "net": {"network": "mobilenet_v4_s", "tag": "il-common"},
    },
)
registry.register_weights(
    "mobilenet_v4_m_il-common",
    {
        "description": "MobileNet v4 medium model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 34.4,
                "sha256": "e229e341117b8c66d13c19d238591665673424358d626de3b33d32e9771310f6",
            }
        },
        "net": {"network": "mobilenet_v4_m", "tag": "il-common"},
    },
)
registry.register_weights(
    "mobilenet_v4_l_il-common",
    {
        "description": "MobileNet v4 large model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 122.0,
                "sha256": "29910ed6bf756648d2887ba0e4f9d8def6231e0ae40978dab92759705db85d1f",
            }
        },
        "net": {"network": "mobilenet_v4_l", "tag": "il-common"},
    },
)
registry.register_weights(
    "mobilenet_v4_l_eu-common",
    {
        "url": "https://huggingface.co/birder-project/mobilenet_v4_l_eu-common/resolve/main",
        "description": "MobileNet v4 large model trained on the eu-common dataset",
        "resolution": (384, 384),
        "formats": {
            "pt": {
                "file_size": 123.7,
                "sha256": "e1a24218a2ee2c5ae3ea7247553ba09fac47bced00dd531a33c7ce06ab49f689",
            }
        },
        "net": {"network": "mobilenet_v4_l", "tag": "eu-common"},
    },
)



import torch
from torch import nn


class LayerScale(nn.Module):
    """
    Layer scale layer that suits 2d in channels last layout or 1d
    """

    def __init__(self, dim: int, init_value: float, inplace: bool = False) -> None:
        super().__init__()
        self.inplace = inplace
        self.gamma = nn.Parameter(init_value * torch.ones(dim))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.inplace is True:
            return x.mul_(self.gamma)

        return x * self.gamma


class LayerScale2d(nn.Module):
    def __init__(self, dim: int, init_value: float, inplace: bool = False) -> None:
        super().__init__()
        self.inplace = inplace
        self.gamma = nn.Parameter(init_value * torch.ones(dim, 1, 1), requires_grad=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.inplace is True:
            return x.mul_(self.gamma)

        return x * self.gamma



"""
XCiT, adapted from
https://github.com/facebookresearch/xcit/blob/main/xcit.py

"XCiT: Cross-Covariance Image Transformers", https://arxiv.org/abs/2106.09681

Changes from original:
* No FPN layers (detection version of XCiT)
"""

# Reference license: Apache-2.0

import math
from collections import OrderedDict
from typing import Any
from typing import Literal
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import MLP
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import StochasticDepth

from birder.common.masking import mask_tensor
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenRetentionResultType
from birder.net.cait import ClassAttention


class SequentialWithSize(nn.Sequential):
    def forward(self, x: torch.Tensor, H: int, W: int) -> torch.Tensor:  # pylint: disable=arguments-differ
        for module in self:
            x = module(x, H, W)

        return x


class PositionalEncodingFourier(nn.Module):
    def __init__(self, hidden_dim: int, dim: int, temperature: int = 10000) -> None:
        super().__init__()
        self.token_projection = nn.Conv2d(hidden_dim * 2, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.scale = 2 * math.pi
        self.temperature = temperature
        self.hidden_dim = hidden_dim

    def forward(self, B: int, H: int, W: int) -> torch.Tensor:
        mask = torch.ones(B, H, W).to(self.token_projection.weight.device)
        y_embed = mask.cumsum(dim=1, dtype=torch.float32)
        x_embed = mask.cumsum(dim=2, dtype=torch.float32)
        eps = 1e-6
        y_embed = y_embed / (y_embed[:, -1:, :] + eps) * self.scale
        x_embed = x_embed / (x_embed[:, :, -1:] + eps) * self.scale

        dim_t = torch.arange(self.hidden_dim, dtype=torch.float32, device=mask.device)
        dim_t = self.temperature ** (2 * (dim_t // 2) / self.hidden_dim)

        pos_x = x_embed[:, :, :, None] / dim_t
        pos_y = y_embed[:, :, :, None] / dim_t
        pos_x = torch.stack((pos_x[:, :, :, 0::2].sin(), pos_x[:, :, :, 1::2].cos()), dim=4).flatten(3)
        pos_y = torch.stack((pos_y[:, :, :, 0::2].sin(), pos_y[:, :, :, 1::2].cos()), dim=4).flatten(3)
        pos = torch.concat((pos_y, pos_x), dim=3).permute(0, 3, 1, 2)
        pos = self.token_projection(pos.to(self.token_projection.weight.dtype))

        return pos


class ConvPatchEmbed(nn.Module):
    def __init__(self, patch_size: Literal[8, 16], in_channels: int, dim: int) -> None:
        super().__init__()
        if patch_size == 16:
            self.proj = nn.Sequential(
                Conv2dNormActivation(
                    in_channels,
                    dim // 8,
                    kernel_size=(3, 3),
                    stride=(2, 2),
                    padding=(1, 1),
                    activation_layer=nn.GELU,
                    inplace=None,
                ),
                Conv2dNormActivation(
                    dim // 8,
                    dim // 4,
                    kernel_size=(3, 3),
                    stride=(2, 2),
                    padding=(1, 1),
                    activation_layer=nn.GELU,
                    inplace=None,
                ),
                Conv2dNormActivation(
                    dim // 4,
                    dim // 2,
                    kernel_size=(3, 3),
                    stride=(2, 2),
                    padding=(1, 1),
                    activation_layer=nn.GELU,
                    inplace=None,
                ),
                Conv2dNormActivation(
                    dim // 2,
                    dim,
                    kernel_size=(3, 3),
                    stride=(2, 2),
                    padding=(1, 1),
                    activation_layer=nn.GELU,
                    inplace=None,
                ),
            )

        elif patch_size == 8:
            self.proj = nn.Sequential(
                Conv2dNormActivation(
                    3,
                    dim // 4,
                    kernel_size=(3, 3),
                    stride=(2, 2),
                    padding=(1, 1),
                    activation_layer=nn.GELU,
                    inplace=None,
                ),
                Conv2dNormActivation(
                    dim // 4,
                    dim // 2,
                    kernel_size=(3, 3),
                    stride=(2, 2),
                    padding=(1, 1),
                    activation_layer=nn.GELU,
                    inplace=None,
                ),
                Conv2dNormActivation(
                    dim // 2,
                    dim,
                    kernel_size=(3, 3),
                    stride=(2, 2),
                    padding=(1, 1),
                    activation_layer=nn.GELU,
                    inplace=None,
                ),
            )

        else:
            raise ValueError("For convolutional projection, patch size has to be in [8, 16]")

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor, int, int]:
        x = self.proj(x)
        H = x.shape[2]
        W = x.shape[3]
        x = x.flatten(2).transpose(1, 2)

        return (x, H, W)


class ClassAttentionBlock(nn.Module):
    def __init__(
        self, dim: int, num_heads: int, mlp_ratio: float, qkv_bias: bool, proj_drop: float, drop_path: float, eta: float
    ) -> None:
        super().__init__()
        self.norm1 = nn.LayerNorm(dim)

        self.attn = ClassAttention(dim, num_heads=num_heads, qkv_bias=qkv_bias, proj_drop=proj_drop)

        self.drop_path = StochasticDepth(drop_path, mode="row")
        self.norm2 = nn.LayerNorm(dim)
        self.mlp = MLP(dim, [int(dim * mlp_ratio), dim], activation_layer=nn.GELU, dropout=proj_drop)

        self.gamma1 = nn.Parameter(eta * torch.ones(dim))
        self.gamma2 = nn.Parameter(eta * torch.ones(dim))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x_norm1 = self.norm1(x)
        x_attn = torch.concat([self.attn(x_norm1), x_norm1[:, 1:]], dim=1)
        x = x + self.drop_path(self.gamma1 * x_attn)
        x = torch.concat([self.norm2(x[:, 0:1]), x[:, 1:]], dim=1)

        x_res = x
        cls_token = x[:, 0:1]
        cls_token = self.gamma2 * self.mlp(cls_token)
        x = torch.concat([cls_token, x[:, 1:]], dim=1)
        x = x_res + self.drop_path(x)

        return x


class LPI(nn.Module):
    """
    Local Patch Interaction module that allows explicit communication between tokens in 3x3 windows to augment the
    implicit communication performed by the block diagonal scatter attention.
    """

    def __init__(self, in_features: int, out_features: int, kernel_size: tuple[int, int]) -> None:
        super().__init__()
        padding = ((kernel_size[0] - 1) // 2, (kernel_size[1] - 1) // 2)

        self.conv_bn_act = Conv2dNormActivation(
            in_features,
            in_features,
            kernel_size=kernel_size,
            stride=(1, 1),
            padding=padding,
            groups=in_features,
            activation_layer=nn.GELU,
            inplace=None,
        )
        self.conv = nn.Conv2d(
            in_features, out_features, kernel_size=kernel_size, stride=(1, 1), padding=padding, groups=out_features
        )

    def forward(self, x: torch.Tensor, H: int, W: int) -> torch.Tensor:
        (B, N, C) = x.shape
        x = x.permute(0, 2, 1).reshape(B, C, H, W)
        x = self.conv_bn_act(x)
        x = self.conv(x)
        x = x.reshape(B, C, N).permute(0, 2, 1)

        return x


class XCA(nn.Module):
    """
    Cross-Covariance Attention (XCA)
    """

    def __init__(self, dim: int, num_heads: int, qkv_bias: bool, proj_drop: float) -> None:
        super().__init__()
        self.num_heads = num_heads
        self.temperature = nn.Parameter(torch.ones(num_heads, 1, 1))

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, N, C) = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads)
        qkv = qkv.permute(2, 0, 3, 1, 4)
        (q, k, v) = qkv.unbind(0)

        q = F.normalize(q, dim=-1) * self.temperature
        k = F.normalize(k, dim=-1)
        x = F.scaled_dot_product_attention(q, k, v, scale=1.0)  # pylint: disable=not-callable

        x = x.permute(0, 3, 1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class XCABlock(nn.Module):
    def __init__(
        self, dim: int, num_heads: int, mlp_ratio: float, qkv_bias: bool, proj_drop: float, drop_path: float, eta: float
    ) -> None:
        super().__init__()
        self.norm1 = nn.LayerNorm(dim)
        self.attn = XCA(dim, num_heads=num_heads, qkv_bias=qkv_bias, proj_drop=proj_drop)
        self.drop_path = StochasticDepth(drop_path, mode="row")

        self.norm2 = nn.LayerNorm(dim)
        self.local_mp = LPI(in_features=dim, out_features=dim, kernel_size=(3, 3))

        self.norm3 = nn.LayerNorm(dim)
        self.mlp = MLP(dim, [int(dim * mlp_ratio), dim], activation_layer=nn.GELU, dropout=proj_drop)

        self.gamma1 = nn.Parameter(eta * torch.ones(dim))
        self.gamma2 = nn.Parameter(eta * torch.ones(dim))
        self.gamma3 = nn.Parameter(eta * torch.ones(dim))

    def forward(self, x: torch.Tensor, H: int, W: int) -> torch.Tensor:
        x = x + self.drop_path(self.gamma1 * self.attn(self.norm1(x)))
        x = x + self.drop_path(self.gamma2 * self.local_mp(self.norm2(x), H, W))
        x = x + self.drop_path(self.gamma3 * self.mlp(self.norm3(x)))

        return x


class XCiT(DetectorBackbone, PreTrainEncoder, MaskedTokenRetentionMixin):
    block_group_regex = r"block1\.stage(\d+)\.(\d+)"  # ClassAttentionBlock combined with the head

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        drop_rate = 0.0
        cls_attn_layers = 2
        mlp_ratio = 4.0
        qkv_bias = True
        patch_size: Literal[8, 16] = self.config["patch_size"]
        embed_dim: int = self.config["embed_dim"]
        depth: int = self.config["depth"]
        num_heads: int = self.config["num_heads"]
        eta: float = self.config["eta"]
        drop_path_rate: float = self.config["drop_path_rate"]

        if depth == 12:
            out_indices = [3, 5, 7, 11]
        elif depth == 24:
            out_indices = [7, 11, 15, 23]
        else:
            raise ValueError(f"depth={depth} is not supported")

        self.patch_embed = ConvPatchEmbed(patch_size, self.input_channels, dim=embed_dim)
        self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, depth)]  # Stochastic depth decay rule

        cur_stage = 0
        block1: OrderedDict[str, nn.Module] = OrderedDict()
        layers = []
        for i in range(depth):
            layers.append(
                XCABlock(
                    dim=embed_dim,
                    num_heads=num_heads,
                    mlp_ratio=mlp_ratio,
                    qkv_bias=qkv_bias,
                    proj_drop=drop_rate,
                    drop_path=dpr[i],
                    eta=eta,
                )
            )
            if i in out_indices:
                block1[f"stage{cur_stage+1}"] = SequentialWithSize(*layers)
                layers = []
                cur_stage += 1

        self.block1 = SequentialWithSize(block1)

        layers2 = []
        for _ in range(cls_attn_layers):
            layers2.append(
                ClassAttentionBlock(
                    dim=embed_dim,
                    num_heads=num_heads,
                    mlp_ratio=mlp_ratio,
                    qkv_bias=qkv_bias,
                    proj_drop=drop_rate,
                    drop_path=0.0,
                    eta=eta,
                )
            )

        layers2.append(nn.LayerNorm(embed_dim))
        self.block2 = nn.Sequential(*layers2)
        self.pos_embed = PositionalEncodingFourier(hidden_dim=32, dim=embed_dim)

        self.return_channels = [embed_dim] * len(out_indices)
        self.embedding_size = embed_dim
        self.classifier = self.create_classifier()

        self.stem_stride = patch_size
        self.stem_width = embed_dim
        self.encoding_size = embed_dim

        # Weights initialization
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            elif isinstance(m, nn.LayerNorm):
                nn.init.zeros_(m.bias)
                nn.init.constant_(m.weight, 1.0)

    def transform_to_backbone(self) -> None:
        self.block2 = nn.Identity()
        self.classifier = nn.Identity()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        B = x.size(0)

        (x, H, W) = self.patch_embed(x)

        pos_encoding = self.pos_embed(B, H, W).reshape(B, -1, x.size(1)).permute(0, 2, 1)
        x = x + pos_encoding

        out = {}
        for name, module in self.block1.named_children():
            x = module(x, H, W)
            if name in self.return_stages:
                out[name] = x.permute(0, 2, 1).reshape(B, -1, H, W)

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.patch_embed.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.block1.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        B = x.size(0)

        (x, H, W) = self.patch_embed(x)
        x = mask_tensor(
            x.permute(0, 2, 1).reshape(B, -1, H, W),
            mask,
            patch_factor=self.max_stride // self.stem_stride,
            mask_token=mask_token,
        )
        x = x.flatten(2).transpose(1, 2)

        pos_encoding = self.pos_embed(B, H, W).reshape(B, -1, x.size(1)).permute(0, 2, 1)
        x = x + pos_encoding
        x = self.block1(x, H, W)

        cls_tokens = self.cls_token.expand(B, -1, -1)
        x = torch.concat((cls_tokens, x), dim=1)
        x = self.block2(x)

        result: TokenRetentionResultType = {}
        if return_keys in ("all", "features"):
            features = x[:, 1:]
            features = features.permute(0, 2, 1)
            (B, C, _) = features.size()
            features = features.reshape(B, C, H, W)
            result["features"] = features

        if return_keys in ("all", "embedding"):
            result["embedding"] = x[:, 0]

        return result

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        B = x.size(0)

        (x, H, W) = self.patch_embed(x)

        pos_encoding = self.pos_embed(B, H, W).reshape(B, -1, x.size(1)).permute(0, 2, 1)
        x = x + pos_encoding
        x = self.block1(x, H, W)

        cls_tokens = self.cls_token.expand(B, -1, -1)
        x = torch.concat((cls_tokens, x), dim=1)
        x = self.block2(x)

        return x

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return x[:, 0]


registry.register_model_config(
    "xcit_nano12_p16",
    XCiT,
    config={"patch_size": 16, "embed_dim": 128, "depth": 12, "num_heads": 4, "eta": 1.0, "drop_path_rate": 0.0},
)
registry.register_model_config(
    "xcit_nano12_p8",
    XCiT,
    config={"patch_size": 8, "embed_dim": 128, "depth": 12, "num_heads": 4, "eta": 1.0, "drop_path_rate": 0.0},
)
registry.register_model_config(
    "xcit_tiny12_p16",
    XCiT,
    config={"patch_size": 16, "embed_dim": 192, "depth": 12, "num_heads": 4, "eta": 1.0, "drop_path_rate": 0.0},
)
registry.register_model_config(
    "xcit_tiny12_p8",
    XCiT,
    config={"patch_size": 8, "embed_dim": 192, "depth": 12, "num_heads": 4, "eta": 1.0, "drop_path_rate": 0.0},
)
registry.register_model_config(
    "xcit_tiny24_p16",
    XCiT,
    config={"patch_size": 16, "embed_dim": 192, "depth": 24, "num_heads": 4, "eta": 1e-5, "drop_path_rate": 0.05},
)
registry.register_model_config(
    "xcit_tiny24_p8",
    XCiT,
    config={"patch_size": 8, "embed_dim": 192, "depth": 24, "num_heads": 4, "eta": 1.0, "drop_path_rate": 0.05},
)
registry.register_model_config(
    "xcit_small12_p16",
    XCiT,
    config={"patch_size": 16, "embed_dim": 384, "depth": 12, "num_heads": 8, "eta": 1.0, "drop_path_rate": 0.05},
)
registry.register_model_config(
    "xcit_small12_p8",
    XCiT,
    config={"patch_size": 8, "embed_dim": 384, "depth": 12, "num_heads": 8, "eta": 1.0, "drop_path_rate": 0.05},
)
registry.register_model_config(
    "xcit_small24_p16",
    XCiT,
    config={"patch_size": 16, "embed_dim": 384, "depth": 24, "num_heads": 8, "eta": 1e-5, "drop_path_rate": 0.1},
)
registry.register_model_config(
    "xcit_small24_p8",
    XCiT,
    config={"patch_size": 8, "embed_dim": 384, "depth": 24, "num_heads": 8, "eta": 1e-5, "drop_path_rate": 0.1},
)
registry.register_model_config(
    "xcit_medium24_p16",
    XCiT,
    config={"patch_size": 16, "embed_dim": 512, "depth": 24, "num_heads": 8, "eta": 1e-5, "drop_path_rate": 0.15},
)
registry.register_model_config(
    "xcit_medium24_p8",
    XCiT,
    config={"patch_size": 8, "embed_dim": 512, "depth": 24, "num_heads": 8, "eta": 1e-5, "drop_path_rate": 0.15},
)
registry.register_model_config(
    "xcit_large24_16",
    XCiT,
    config={"patch_size": 16, "embed_dim": 768, "depth": 24, "num_heads": 16, "eta": 1e-5, "drop_path_rate": 0.25},
)
registry.register_model_config(
    "xcit_large24_8",
    XCiT,
    config={"patch_size": 8, "embed_dim": 768, "depth": 24, "num_heads": 16, "eta": 1e-5, "drop_path_rate": 0.3},
)

registry.register_weights(
    "xcit_nano12_p16_il-common",
    {
        "url": "https://huggingface.co/birder-project/xcit_nano12_p16_il-common/resolve/main",
        "description": "XCiT nano d12 patch16 model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 11.5,
                "sha256": "f0b1d9b6a39f816b235795155a9afd3fa344ba95bb083d7509d5f4e4d93dde3c",
            },
        },
        "net": {"network": "xcit_nano12_p16", "tag": "il-common"},
    },
)
registry.register_weights(
    "xcit_nano12_p8_il-common",
    {
        "description": "XCiT nano d12 patch8 model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 11.5,
                "sha256": "fe14249630e41102ad3059ffa1afe581d659603979f1bab0a1a32cbdb95d3b83",
            },
        },
        "net": {"network": "xcit_nano12_p8", "tag": "il-common"},
    },
)



"""
Adapted from https://github.com/jacobgil/pytorch-grad-cam
"""

# Reference license: MIT

from collections.abc import Callable
from pathlib import Path
from typing import Optional

import numpy as np
import numpy.typing as npt
import torch
from PIL import Image
from torch import nn
from torch.utils.hooks import RemovableHandle

from birder.introspection.base import InterpretabilityResult
from birder.introspection.base import Interpreter
from birder.introspection.base import show_mask_on_image


def _scale_cam_image(
    cam: npt.NDArray[np.float32], target_size: Optional[tuple[int, int]] = None
) -> npt.NDArray[np.float32]:
    result = []
    for img in cam:
        img = img - np.min(img)
        img = img / (1e-7 + np.max(img))
        if target_size is not None:
            img = np.array(Image.fromarray(img).resize(target_size))

        result.append(img)

    return np.array(result, dtype=np.float32)


class ClassifierOutputTarget:
    def __init__(self, category: int) -> None:
        self.category = category

    def __call__(self, model_output: torch.Tensor) -> torch.Tensor:
        if len(model_output.shape) == 1:
            return model_output[self.category]

        return model_output[:, self.category]


class ActivationsAndGradients:
    """
    Class for extracting activations and
    registering gradients from targeted intermediate layers
    """

    def __init__(
        self,
        model: nn.Module,
        target_layer: nn.Module,
        reshape_transform: Optional[Callable[[torch.Tensor], torch.Tensor]],
    ) -> None:
        self.model = model
        self.gradients: torch.Tensor
        self.activations: torch.Tensor
        self.reshape_transform = reshape_transform
        self.handles: list[RemovableHandle] = []

        self.handles.append(target_layer.register_forward_hook(self.save_activation))
        # Because of https://github.com/pytorch/pytorch/issues/61519,
        # we don't use backward hook to record gradients.
        self.handles.append(target_layer.register_forward_hook(self.save_gradient))

    def save_activation(self, _module: nn.Module, _input: torch.Tensor, output: torch.Tensor) -> None:
        if self.reshape_transform is not None:
            output = self.reshape_transform(output)

        self.activations = output.cpu().detach()

    def save_gradient(self, _module: nn.Module, _input: torch.Tensor, output: torch.Tensor) -> None:
        if hasattr(output, "requires_grad") is False or output.requires_grad is False:
            # You can only register hooks on tensor requires grad.
            return

        # Gradients are computed in reverse order
        def _store_grad(grad: torch.Tensor) -> None:
            if self.reshape_transform is not None:
                grad = self.reshape_transform(grad)

            self.gradients = grad.cpu().detach()

        output.register_hook(_store_grad)

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return self.model(x)

    def release(self) -> None:
        for handle in self.handles:
            handle.remove()


class GradCAM:
    def __init__(
        self,
        model: nn.Module,
        target_layer: nn.Module,
        reshape_transform: Optional[Callable[[torch.Tensor], torch.Tensor]] = None,
    ) -> None:
        self.model = model.eval()
        self.target_layer = target_layer
        self.activations_and_grads = ActivationsAndGradients(self.model, target_layer, reshape_transform)

    def get_cam_image(
        self, activations: npt.NDArray[np.float32], grads: npt.NDArray[np.float32]
    ) -> npt.NDArray[np.float32]:
        weights: npt.NDArray[np.float32] = np.mean(grads, axis=(2, 3))
        weighted_activations = weights[:, :, None, None] * activations
        cam: npt.NDArray[np.float32] = weighted_activations.sum(axis=1)

        return cam

    def compute_layer_cam(self, input_tensor: torch.Tensor) -> npt.NDArray[np.float32]:
        target_size = (input_tensor.size(-1), input_tensor.size(-2))

        layer_activations = self.activations_and_grads.activations.numpy()
        layer_grads = self.activations_and_grads.gradients.numpy()

        cam = self.get_cam_image(layer_activations, layer_grads)
        cam = np.maximum(cam, 0)
        scaled = _scale_cam_image(cam, target_size)
        return scaled[:, None, :]

    def __call__(
        self, input_tensor: torch.Tensor, target: Optional[ClassifierOutputTarget] = None
    ) -> npt.NDArray[np.float32]:
        output = self.activations_and_grads(input_tensor)
        if target is None:
            category = np.argmax(output.cpu().data.numpy(), axis=-1)
            target = ClassifierOutputTarget(category)

        self.model.zero_grad()
        loss = target(output)
        loss.backward(retain_graph=True)

        cam_per_layer = self.compute_layer_cam(input_tensor)
        cam_per_layer = np.mean(cam_per_layer, axis=1)
        cam_per_layer = _scale_cam_image(cam_per_layer)

        return cam_per_layer

    def __del__(self) -> None:
        self.activations_and_grads.release()


class GradCamInterpreter(Interpreter):
    def __init__(
        self,
        model: nn.Module,
        device: torch.device,
        transform: Callable[..., torch.Tensor],
        target_layer: nn.Module,
        reshape_transform: Optional[Callable[[torch.Tensor], torch.Tensor]] = None,
    ) -> None:
        super().__init__(model, device, transform)
        self.grad_cam = GradCAM(model, target_layer, reshape_transform=reshape_transform)

    def interpret(self, image: str | Path | Image.Image, target_class: Optional[int] = None) -> InterpretabilityResult:
        (input_tensor, rgb_img) = self._preprocess_image(image)

        if target_class is not None:
            target = ClassifierOutputTarget(target_class)
        else:
            target = None

        grayscale_cam = self.grad_cam(input_tensor, target=target)[0, :]
        visualization = show_mask_on_image(rgb_img, grayscale_cam)

        return InterpretabilityResult(rgb_img, visualization, raw_output=grayscale_cam)
